# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Model definitions for simple speech recognition.

Brainchip information:
        - Original file: https://github.com/tensorflow/tensorflow/tree/master/tensorflow/examples/speech_commands
"""
import hashlib
import math
import os.path
import random
import re
import sys
import tarfile
import glob

import numpy as np
import urllib
import tensorflow as tf

from tensorflow.python.ops import gen_audio_ops as audio_ops

MAX_NUM_WAVS_PER_CLASS = 2**27 - 1  # ~134M
SILENCE_LABEL = '_silence_'
SILENCE_INDEX = 0
UNKNOWN_WORD_LABEL = '_unknown_'
UNKNOWN_WORD_INDEX = 1
BACKGROUND_NOISE_DIR_NAME = '_background_noise_'
RANDOM_SEED = 59185


def prepare_model_settings(sample_rate, clip_duration_ms, window_size_ms,
                           window_stride_ms, feature_bin_count):
    """Calculates common settings needed for all models.

  Args:
    sample_rate: Number of audio samples per second.
    clip_duration_ms: Length of each audio clip to be analyzed.
    window_size_ms: Duration of frequency analysis window.
    window_stride_ms: How far to move in time between frequency windows.
    feature_bin_count: Number of frequency bins to use for analysis.

  Returns:
    Dictionary containing common settings.

  Raises:
    ValueError: If the preprocessing mode isn't recognized.
  """
    desired_samples = int(sample_rate * clip_duration_ms / 1000)
    window_size_samples = int(sample_rate * window_size_ms / 1000)
    window_stride_samples = int(sample_rate * window_stride_ms / 1000)
    length_minus_window = (desired_samples - window_size_samples)
    if length_minus_window < 0:
        spectrogram_length = 0
    else:
        spectrogram_length = 1 + int(
            length_minus_window / window_stride_samples)
    average_window_width = -1
    fingerprint_width = feature_bin_count

    fingerprint_size = fingerprint_width * spectrogram_length
    return {
        'desired_samples': desired_samples,
        'window_size_samples': window_size_samples,
        'window_stride_samples': window_stride_samples,
        'spectrogram_length': spectrogram_length,
        'fingerprint_width': fingerprint_width,
        'fingerprint_size': fingerprint_size,
    }


def prepare_words_list(wanted_words):
    """Prepends common tokens to the custom word list.

  Args:
    wanted_words: List of strings containing the custom words.

  Returns:
    List with the standard silence and unknown tokens added.
  """
    return [SILENCE_LABEL, UNKNOWN_WORD_LABEL] + wanted_words


def which_set(filename, validation_percentage, testing_percentage):
    """Determines which data partition the file should belong to.

  We want to keep files in the same training, validation, or testing sets even
  if new ones are added over time. This makes it less likely that testing
  samples will accidentally be reused in training when long runs are restarted
  for example. To keep this stability, a hash of the filename is taken and used
  to determine which set it should belong to. This determination only depends on
  the name and the set proportions, so it won't change as other files are added.

  It's also useful to associate particular files as related (for example words
  spoken by the same person), so anything after '_nohash_' in a filename is
  ignored for set determination. This ensures that 'bobby_nohash_0.wav' and
  'bobby_nohash_1.wav' are always in the same set, for example.

  Args:
    filename: File path of the data sample.
    validation_percentage: How much of the data set to use for validation.
    testing_percentage: How much of the data set to use for testing.

  Returns:
    String, one of 'training', 'validation', or 'testing'.
  """
    base_name = os.path.basename(filename)
    # We want to ignore anything after '_nohash_' in the file name when
    # deciding which set to put a wav in, so the data set creator has a way of
    # grouping wavs that are close variations of each other.
    hash_name = re.sub(r'_nohash_.*$', '', base_name)
    # This looks a bit magical, but we need to decide whether this file should
    # go into the training, testing, or validation sets, and we want to keep
    # existing files in the same set even if more files are subsequently
    # added.
    # To do that, we need a stable way of deciding based on just the file name
    # itself, so we do a hash of that and then use that to generate a
    # probability value that we use to assign it.
    hash_name_hashed = hashlib.sha1(tf.compat.as_bytes(hash_name)).hexdigest()
    percentage_hash = ((int(hash_name_hashed, 16) %
                        (MAX_NUM_WAVS_PER_CLASS + 1)) *
                       (100.0 / MAX_NUM_WAVS_PER_CLASS))
    if percentage_hash < validation_percentage:
        result = 'validation'
    elif percentage_hash < (testing_percentage + validation_percentage):
        result = 'testing'
    else:
        result = 'training'
    return result


class AudioProcessor(object):
    """Handles loading, partitioning, and preparing audio training data."""

    def __init__(self,
                 sample_rate,
                 clip_duration_ms,
                 window_size_ms,
                 window_stride_ms,
                 feature_bin_count,
                 data_url=None,
                 data_dir=None,
                 silence_percentage=0,
                 unknown_percentage=0,
                 wanted_words=None,
                 validation_percentage=0,
                 testing_percentage=0):
        self.data_dir = None
        self.background_data = None
        self.model_settings = prepare_model_settings(sample_rate,
                                                     clip_duration_ms,
                                                     window_size_ms,
                                                     window_stride_ms,
                                                     feature_bin_count)
        if data_dir:
            self.data_dir = data_dir
            self.maybe_download_and_extract_dataset(data_url, data_dir)
            self.prepare_data_index(silence_percentage, unknown_percentage,
                                    wanted_words, validation_percentage,
                                    testing_percentage)
            self.prepare_background_data()
        self.prepare_processing_graph()

    def maybe_download_and_extract_dataset(self, data_url, dest_directory):
        """Download and extract data set tar file.

    If the data set we're using doesn't already exist, this function
    downloads it from the TensorFlow.org website and unpacks it into a
    directory.
    If the data_url is none, don't download anything and expect the data
    directory to contain the correct files already.

    Args:
      data_url: Web location of the tar file containing the data set.
      dest_directory: File path to extract data to.
    """
        if not data_url:
            return
        if not os.path.exists(dest_directory):
            os.makedirs(dest_directory)
        filename = data_url.split('/')[-1]
        filepath = os.path.join(dest_directory, filename)
        if not os.path.exists(filepath):

            def _progress(count, block_size, total_size):
                sys.stdout.write('\r>> Downloading %s %.1f%%' %
                                 (filename, float(count * block_size) /
                                  float(total_size) * 100.0))
                sys.stdout.flush()

            try:
                filepath, _ = urllib.request.urlretrieve(
                    data_url, filepath, _progress)
            except:
                raise RuntimeError(
                    f"Failed to download URL: {data_url} to folder: "
                    f"{filepath}\nPlease make sure you have enough free"
                    f" space and an internet connection")
            print()
            statinfo = os.stat(filepath)
            print(
                f"Successfully downloaded {filename} ({statinfo.st_size} bytes)"
            )
        tarfile.open(filepath, 'r:gz').extractall(dest_directory)

    def prepare_data_index(self, silence_percentage, unknown_percentage,
                           wanted_words, validation_percentage,
                           testing_percentage):
        """Prepares a list of the samples organized by set and label.

    The training loop needs a list of all the available data, organized by
    which partition it should belong to, and with ground truth labels attached.
    This function analyzes the folders below the `data_dir`, figures out the
    right
    labels for each file based on the name of the subdirectory it belongs to,
    and uses a stable hash to assign it to a data set partition.

    Args:
      silence_percentage: How much of the resulting data should be background.
      unknown_percentage: How much should be audio outside the wanted classes.
      wanted_words: Labels of the classes we want to be able to recognize.
      validation_percentage: How much of the data set to use for validation.
      testing_percentage: How much of the data set to use for testing.

    Returns:
      Dictionary containing a list of file information for each set partition,
      and a lookup map for each class to determine its numeric index.

    Raises:
      Exception: If expected files are not found.
    """
        # Make sure the shuffling and picking of unknowns is deterministic.
        random.seed(RANDOM_SEED)
        wanted_words_index = {}
        for index, wanted_word in enumerate(wanted_words):
            wanted_words_index[wanted_word] = index + 2
        self.data_index = {'validation': [], 'testing': [], 'training': []}
        unknown_index = {'validation': [], 'testing': [], 'training': []}
        all_words = {}
        # Look through all the subfolders to find audio samples
        search_path = os.path.join(self.data_dir, '*', '*.wav')
        for wav_path in glob.glob(search_path):
            _, word = os.path.split(os.path.dirname(wav_path))
            word = word.lower()
            # Treat the '_background_noise_' folder as a special case, since we expect
            # it to contain long audio samples we mix in to improve training.
            if word == BACKGROUND_NOISE_DIR_NAME:
                continue
            all_words[word] = True
            set_index = which_set(wav_path, validation_percentage,
                                  testing_percentage)
            # If it's a known class, store its detail, otherwise add it to the list
            # we'll use to train the unknown label.
            if word in wanted_words_index:
                self.data_index[set_index].append({
                    'label': word,
                    'file': wav_path
                })
            else:
                unknown_index[set_index].append({
                    'label': word,
                    'file': wav_path
                })
        if not all_words:
            raise Exception('No .wavs found at ' + search_path)
        for index, wanted_word in enumerate(wanted_words):
            if wanted_word not in all_words:
                raise Exception('Expected to find ' + wanted_word +
                                ' in labels but only found ' +
                                ', '.join(all_words.keys()))
        # We need an arbitrary file to load as the input for the silence samples.
        # It's multiplied by zero later, so the content doesn't matter.
        silence_wav_path = self.data_index['training'][0]['file']
        for set_index in ['validation', 'testing', 'training']:
            set_size = len(self.data_index[set_index])
            silence_size = int(math.ceil(set_size * silence_percentage / 100))
            for _ in range(silence_size):
                self.data_index[set_index].append({
                    'label': SILENCE_LABEL,
                    'file': silence_wav_path
                })
            # Pick some unknowns to add to each partition of the data set.
            random.shuffle(unknown_index[set_index])
            unknown_size = int(math.ceil(set_size * unknown_percentage / 100))
            self.data_index[set_index].extend(
                unknown_index[set_index][:unknown_size])
        # Make sure the ordering is random.
        for set_index in ['validation', 'testing', 'training']:
            random.shuffle(self.data_index[set_index])
        # Prepare the rest of the result data structure.
        self.words_list = prepare_words_list(wanted_words)
        self.word_to_index = {}
        for word in all_words:
            if word in wanted_words_index:
                self.word_to_index[word] = wanted_words_index[word]
            else:
                self.word_to_index[word] = UNKNOWN_WORD_INDEX
        self.word_to_index[SILENCE_LABEL] = SILENCE_INDEX

    def prepare_background_data(self):
        """Searches a folder for background noise audio, and loads it into memory.

    It's expected that the background audio samples will be in a subdirectory
    named '_background_noise_' inside the 'data_dir' folder, as .wavs that match
    the sample rate of the training data, but can be much longer in duration.

    If the '_background_noise_' folder doesn't exist at all, this isn't an
    error, it's just taken to mean that no background noise augmentation should
    be used. If the folder does exist, but it's empty, that's treated as an
    error.

    Returns:
      List of raw PCM-encoded audio samples of background noise.

    Raises:
      Exception: If files aren't found in the folder.
    """
        self.background_data = []
        background_dir = os.path.join(self.data_dir, BACKGROUND_NOISE_DIR_NAME)
        if not os.path.exists(background_dir):
            return self.background_data

        search_path = os.path.join(self.data_dir, BACKGROUND_NOISE_DIR_NAME,
                                   '*.wav')
        for wav_path in glob.glob(search_path):
            wav_loader = tf.io.read_file(wav_path)
            wav_decoder = tf.audio.decode_wav(wav_loader, desired_channels=1)
            wav_data = wav_decoder.audio.numpy().flatten()
            self.background_data.append(wav_data)
        if not self.background_data:
            raise Exception('No background wav files were found in ' +
                            search_path)

    def prepare_processing_graph(self):
        """Builds a TensorFlow graph to apply the input distortions.

    Creates a graph that loads a WAVE file, decodes it, scales the volume,
    shifts it in time, adds in background noise, calculates a spectrogram, and
    then builds an MFCC fingerprint from that.
    """

        @tf.function
        def processing_graph(wav_filename, foreground_volume,
                             time_shift_padding, time_shift_offset,
                             background_data, background_volume,
                             model_settings):
            desired_samples = model_settings['desired_samples']
            wav_loader = tf.io.read_file(wav_filename)
            wav_decoder = tf.audio.decode_wav(wav_loader,
                                              desired_channels=1,
                                              desired_samples=desired_samples)
            # Allow the audio sample's volume to be adjusted.
            scaled_foreground = tf.multiply(wav_decoder.audio,
                                            foreground_volume)
            # Shift the sample's start position, and pad any gaps with zeros.
            padded_foreground = tf.pad(tensor=scaled_foreground,
                                       paddings=time_shift_padding,
                                       mode='CONSTANT')
            sliced_foreground = tf.slice(padded_foreground, time_shift_offset,
                                         [desired_samples, -1])
            # Mix in background noise.
            background_mul = tf.multiply(background_data, background_volume)
            background_add = tf.add(background_mul, sliced_foreground)
            background_clamp = tf.clip_by_value(background_add, -1.0, 1.0)
            # Run the spectrogram and MFCC ops to get a 2D 'fingerprint' of the audio.
            spectrogram = audio_ops.audio_spectrogram(
                background_clamp,
                window_size=model_settings['window_size_samples'],
                stride=model_settings['window_stride_samples'],
                magnitude_squared=True)
            # The number of buckets in each FFT row in the spectrogram will depend on
            # how many input samples there are in each window. This can be quite
            # large, with a 160 sample window producing 127 buckets for example. We
            # don't need this level of detail for classification, so we often want to
            # shrink them down to produce a smaller result. That's what this section
            # implements. One method is to use average pooling to merge adjacent
            # buckets, but a more sophisticated approach is to apply the MFCC
            # algorithm to shrink the representation.
            output = audio_ops.mfcc(
                spectrogram,
                wav_decoder.sample_rate,
                dct_coefficient_count=model_settings['fingerprint_width'])
            return output

        self.processing_graph = processing_graph

    def get_data(self, how_many, offset, background_frequency,
                 background_volume_range, time_shift, mode):
        """Gather samples from the data set, applying transformations as needed.

    When the mode is 'training', a random selection of samples will be returned,
    otherwise the first N clips in the partition will be used. This ensures that
    validation always uses the same samples, reducing noise in the metrics.

    Args:
      how_many: Desired number of samples to return. -1 means the entire
        contents of this partition.
      offset: Where to start when fetching deterministically.
      background_frequency: How many clips will have background noise, 0.0 to
        1.0.
      background_volume_range: How loud the background noise will be.
      time_shift: How much to randomly shift the clips by in time.
      mode: Which partition to use, must be 'training', 'validation', or
        'testing'.

    Returns:
      List of sample data for the transformed samples, and list of label indexes

    Raises:
      ValueError: If background samples are too short.
    """
        # Raise an error if data are not present, i.e. data directory is None
        if self.data_dir is None:
            raise RuntimeError(
                "Calling 'get_data' method requires to have a valid "
                "data directory. Here, data_dir=None")

        # Pick one of the partitions to choose samples from.
        candidates = self.data_index[mode]
        if how_many == -1:
            sample_count = len(candidates)
        else:
            sample_count = max(0, min(how_many, len(candidates) - offset))
        # Data and labels will be populated and returned.
        data = np.zeros((sample_count, self.model_settings['fingerprint_size']))
        labels = np.zeros(sample_count)
        desired_samples = self.model_settings['desired_samples']
        use_background = self.background_data and (mode == 'training')
        pick_deterministically = (mode != 'training')
        # Use the processing graph we created earlier to repeatedly to generate the
        # final output sample data we'll use in training.
        for i in range(offset, offset + sample_count):
            # Pick which audio sample to use.
            if how_many == -1 or pick_deterministically:
                sample_index = i
            else:
                sample_index = np.random.randint(len(candidates))
            sample = candidates[sample_index]
            # If we're time shifting, set up the offset for this sample.
            if time_shift > 0:
                time_shift_amount = np.random.randint(-time_shift, time_shift)
            else:
                time_shift_amount = 0
            if time_shift_amount > 0:
                time_shift_padding = [[time_shift_amount, 0], [0, 0]]
                time_shift_offset = [0, 0]
            else:
                time_shift_padding = [[0, -time_shift_amount], [0, 0]]
                time_shift_offset = [-time_shift_amount, 0]
            input_dict = {
                'wav_filename': tf.constant(sample['file']),
                'time_shift_padding': tf.constant(time_shift_padding),
                'time_shift_offset': tf.constant(time_shift_offset)
            }
            # Choose a section of background noise to mix in.
            if use_background or sample['label'] == SILENCE_LABEL:
                background_index = np.random.randint(len(self.background_data))
                background_samples = self.background_data[background_index]
                if len(background_samples
                      ) <= self.model_settings['desired_samples']:
                    raise ValueError(
                        'Background sample is too short! Need more than %d'
                        ' samples but only %d were found' %
                        (self.model_settings['desired_samples'],
                         len(background_samples)))
                background_offset = np.random.randint(
                    0,
                    len(background_samples) -
                    self.model_settings['desired_samples'])
                background_clipped = background_samples[background_offset:(
                    background_offset + desired_samples)]
                background_reshaped = background_clipped.reshape(
                    [desired_samples, 1])
                if sample['label'] == SILENCE_LABEL:
                    background_volume = np.random.uniform(0, 1)
                elif np.random.uniform(0, 1) < background_frequency:
                    background_volume = np.random.uniform(
                        0, background_volume_range)
                else:
                    background_volume = 0
            else:
                background_reshaped = np.zeros([desired_samples, 1])
                background_volume = 0
            input_dict['background_data'] = tf.constant(background_reshaped,
                                                        dtype=tf.float32)
            input_dict['background_volume'] = tf.constant(background_volume,
                                                          dtype=tf.float32)
            # If we want silence, mute out the main sample but leave the background.
            if sample['label'] == SILENCE_LABEL:
                input_dict['foreground_volume'] = tf.constant(0.0)
            else:
                input_dict['foreground_volume'] = tf.constant(1.0)
            # Run the graph to produce the output audio.
            data_tensor = self.processing_graph(
                **input_dict, model_settings=self.model_settings)
            data[i - offset, :] = data_tensor.numpy().flatten()
            label_index = self.word_to_index[sample['label']]
            labels[i - offset] = label_index
        return data, labels

    def get_features_for_wav(self, wav_filename):
        """Applies the feature transformation process to the input_wav.

    Runs the feature generation process (generally producing a spectrogram from
    the input samples) on the WAV file. This can be useful for testing and
    verifying implementations being run on other platforms.

    Args:
      wav_filename: The path to the input audio file.

    Returns:
      Numpy data array containing the generated features.
    """
        desired_samples = self.model_settings['desired_samples']
        background_data = tf.constant(np.zeros([desired_samples, 1]),
                                      tf.float32)
        input_dict = {
            'wav_filename': tf.constant(wav_filename),
            'time_shift_padding': tf.constant([[0, 0], [0, 0]]),
            'time_shift_offset': tf.constant([0, 0]),
            'background_data': background_data,
            'background_volume': tf.constant(0.0),
            'foreground_volume': tf.constant(1.0)
        }
        # Run the graph to produce the output audio.
        data_tensor = self.processing_graph(**input_dict,
                                            model_settings=self.model_settings)
        features = data_tensor.numpy()
        features = features.reshape(
            (self.model_settings['spectrogram_length'],
             self.model_settings['fingerprint_width'], 1))
        return features

    def get_augmented_data_for_wav(self,
                                   wav_filename,
                                   background_frequency,
                                   background_volume_range,
                                   time_shift,
                                   num_augmented_samples=1):
        """Applies the feature transformation process to a wav audio file, adding
    data augmentation (background noise and time shifting).

    Args:
      wav_filename (str): The path to the input audio file.
      background_frequency: How many clips will have background noise, 0.0 to
        1.0.
      background_volume_range: How loud the background noise will be.
      time_shift: How much to randomly shift the clips by in time.
      num_augmented_samples: How many samples will be generated using data
        augmentation.

    Returns:
      Numpy data array containing the generated features for every augmented
        sample.
    """
        data = np.zeros(
            (num_augmented_samples, self.model_settings['fingerprint_size']))
        desired_samples = self.model_settings['desired_samples']

        for i in range(num_augmented_samples):
            # If we're time shifting, set up the offset for this sample.
            if time_shift > 0:
                time_shift_amount = np.random.randint(-time_shift, time_shift)
            else:
                time_shift_amount = 0
            if time_shift_amount > 0:
                time_shift_padding = [[time_shift_amount, 0], [0, 0]]
                time_shift_offset = [0, 0]
            else:
                time_shift_padding = [[0, -time_shift_amount], [0, 0]]
                time_shift_offset = [-time_shift_amount, 0]
            input_dict = {
                'wav_filename': tf.constant(wav_filename),
                'foreground_volume': tf.constant(1.0),
                'time_shift_padding': tf.constant(time_shift_padding),
                'time_shift_offset': tf.constant(time_shift_offset)
            }

            # Choose a section of background noise to mix in.
            if self.background_data:
                background_index = np.random.randint(len(self.background_data))
                background_samples = self.background_data[background_index]
                if len(background_samples
                      ) <= self.model_settings['desired_samples']:
                    raise ValueError(
                        'Background sample is too short! Need more than %d'
                        ' samples but only %d were found' %
                        (self.model_settings['desired_samples'],
                         len(background_samples)))
                background_offset = np.random.randint(
                    0,
                    len(background_samples) -
                    self.model_settings['desired_samples'])
                background_clipped = background_samples[background_offset:(
                    background_offset + desired_samples)]
                background_reshaped = background_clipped.reshape(
                    [desired_samples, 1])
                if np.random.uniform(0, 1) < background_frequency:
                    background_volume = np.random.uniform(
                        0, background_volume_range)
                else:
                    background_volume = 0
            else:
                background_reshaped = np.zeros([desired_samples, 1])
                background_volume = 0
            input_dict['background_data'] = tf.constant(background_reshaped,
                                                        dtype=tf.float32)
            input_dict['background_volume'] = tf.constant(background_volume,
                                                          dtype=tf.float32)

            data_tensor = self.processing_graph(
                **input_dict, model_settings=self.model_settings)
            data[i, :] = data_tensor.numpy().flatten()

        return data



imagenet_labels = {
    0:
        'tench, Tinca tinca',
    1:
        'goldfish, Carassius auratus',
    2:
        'great white shark, white shark, man-eater, man-eating shark, Carcharodon carcharias',
    3:
        'tiger shark, Galeocerdo cuvieri',
    4:
        'hammerhead, hammerhead shark',
    5:
        'electric ray, crampfish, numbfish, torpedo',
    6:
        'stingray',
    7:
        'cock',
    8:
        'hen',
    9:
        'ostrich, Struthio camelus',
    10:
        'brambling, Fringilla montifringilla',
    11:
        'goldfinch, Carduelis carduelis',
    12:
        'house finch, linnet, Carpodacus mexicanus',
    13:
        'junco, snowbird',
    14:
        'indigo bunting, indigo finch, indigo bird, Passerina cyanea',
    15:
        'robin, American robin, Turdus migratorius',
    16:
        'bulbul',
    17:
        'jay',
    18:
        'magpie',
    19:
        'chickadee',
    20:
        'water ouzel, dipper',
    21:
        'kite',
    22:
        'bald eagle, American eagle, Haliaeetus leucocephalus',
    23:
        'vulture',
    24:
        'great grey owl, great gray owl, Strix nebulosa',
    25:
        'European fire salamander, Salamandra salamandra',
    26:
        'common newt, Triturus vulgaris',
    27:
        'eft',
    28:
        'spotted salamander, Ambystoma maculatum',
    29:
        'axolotl, mud puppy, Ambystoma mexicanum',
    30:
        'bullfrog, Rana catesbeiana',
    31:
        'tree frog, tree-frog',
    32:
        'tailed frog, bell toad, ribbed toad, tailed toad, Ascaphus trui',
    33:
        'loggerhead, loggerhead turtle, Caretta caretta',
    34:
        'leatherback turtle, leatherback, leathery turtle, Dermochelys coriacea',
    35:
        'mud turtle',
    36:
        'terrapin',
    37:
        'box turtle, box tortoise',
    38:
        'banded gecko',
    39:
        'common iguana, iguana, Iguana iguana',
    40:
        'American chameleon, anole, Anolis carolinensis',
    41:
        'whiptail, whiptail lizard',
    42:
        'agama',
    43:
        'frilled lizard, Chlamydosaurus kingi',
    44:
        'alligator lizard',
    45:
        'Gila monster, Heloderma suspectum',
    46:
        'green lizard, Lacerta viridis',
    47:
        'African chameleon, Chamaeleo chamaeleon',
    48:
        'Komodo dragon, Komodo lizard, dragon lizard, giant lizard, Varanus komodoensis',
    49:
        'African crocodile, Nile crocodile, Crocodylus niloticus',
    50:
        'American alligator, Alligator mississipiensis',
    51:
        'triceratops',
    52:
        'thunder snake, worm snake, Carphophis amoenus',
    53:
        'ringneck snake, ring-necked snake, ring snake',
    54:
        'hognose snake, puff adder, sand viper',
    55:
        'green snake, grass snake',
    56:
        'king snake, kingsnake',
    57:
        'garter snake, grass snake',
    58:
        'water snake',
    59:
        'vine snake',
    60:
        'night snake, Hypsiglena torquata',
    61:
        'boa constrictor, Constrictor constrictor',
    62:
        'rock python, rock snake, Python sebae',
    63:
        'Indian cobra, Naja naja',
    64:
        'green mamba',
    65:
        'sea snake',
    66:
        'horned viper, cerastes, sand viper, horned asp, Cerastes cornutus',
    67:
        'diamondback, diamondback rattlesnake, Crotalus adamanteus',
    68:
        'sidewinder, horned rattlesnake, Crotalus cerastes',
    69:
        'trilobite',
    70:
        'harvestman, daddy longlegs, Phalangium opilio',
    71:
        'scorpion',
    72:
        'black and gold garden spider, Argiope aurantia',
    73:
        'barn spider, Araneus cavaticus',
    74:
        'garden spider, Aranea diademata',
    75:
        'black widow, Latrodectus mactans',
    76:
        'tarantula',
    77:
        'wolf spider, hunting spider',
    78:
        'tick',
    79:
        'centipede',
    80:
        'black grouse',
    81:
        'ptarmigan',
    82:
        'ruffed grouse, partridge, Bonasa umbellus',
    83:
        'prairie chicken, prairie grouse, prairie fowl',
    84:
        'peacock',
    85:
        'quail',
    86:
        'partridge',
    87:
        'African grey, African gray, Psittacus erithacus',
    88:
        'macaw',
    89:
        'sulphur-crested cockatoo, Kakatoe galerita, Cacatua galerita',
    90:
        'lorikeet',
    91:
        'coucal',
    92:
        'bee eater',
    93:
        'hornbill',
    94:
        'hummingbird',
    95:
        'jacamar',
    96:
        'toucan',
    97:
        'drake',
    98:
        'red-breasted merganser, Mergus serrator',
    99:
        'goose',
    100:
        'black swan, Cygnus atratus',
    101:
        'tusker',
    102:
        'echidna, spiny anteater, anteater',
    103:
        'platypus, duckbill, duckbilled platypus, duck-billed platypus, Ornithorhynchus anatinus',
    104:
        'wallaby, brush kangaroo',
    105:
        'koala, koala bear, kangaroo bear, native bear, Phascolarctos cinereus',
    106:
        'wombat',
    107:
        'jellyfish',
    108:
        'sea anemone, anemone',
    109:
        'brain coral',
    110:
        'flatworm, platyhelminth',
    111:
        'nematode, nematode worm, roundworm',
    112:
        'conch',
    113:
        'snail',
    114:
        'slug',
    115:
        'sea slug, nudibranch',
    116:
        'chiton, coat-of-mail shell, sea cradle, polyplacophore',
    117:
        'chambered nautilus, pearly nautilus, nautilus',
    118:
        'Dungeness crab, Cancer magister',
    119:
        'rock crab, Cancer irroratus',
    120:
        'fiddler crab',
    121:
        'king crab, Alaska crab, Alaskan king crab, Alaska king crab, Paralithodes camtschatica',
    122:
        'American lobster, Northern lobster, Maine lobster, Homarus americanus',
    123:
        'spiny lobster, langouste, rock lobster, crawfish, crayfish, sea crawfish',
    124:
        'crayfish, crawfish, crawdad, crawdaddy',
    125:
        'hermit crab',
    126:
        'isopod',
    127:
        'white stork, Ciconia ciconia',
    128:
        'black stork, Ciconia nigra',
    129:
        'spoonbill',
    130:
        'flamingo',
    131:
        'little blue heron, Egretta caerulea',
    132:
        'American egret, great white heron, Egretta albus',
    133:
        'bittern',
    134:
        'crane',
    135:
        'limpkin, Aramus pictus',
    136:
        'European gallinule, Porphyrio porphyrio',
    137:
        'American coot, marsh hen, mud hen, water hen, Fulica americana',
    138:
        'bustard',
    139:
        'ruddy turnstone, Arenaria interpres',
    140:
        'red-backed sandpiper, dunlin, Erolia alpina',
    141:
        'redshank, Tringa totanus',
    142:
        'dowitcher',
    143:
        'oystercatcher, oyster catcher',
    144:
        'pelican',
    145:
        'king penguin, Aptenodytes patagonica',
    146:
        'albatross, mollymawk',
    147:
        'grey whale, gray whale, devilfish, Eschrichtius gibbosus, Eschrichtius robustus',
    148:
        'killer whale, killer, orca, grampus, sea wolf, Orcinus orca',
    149:
        'dugong, Dugong dugon',
    150:
        'sea lion',
    151:
        'Chihuahua',
    152:
        'Japanese spaniel',
    153:
        'Maltese dog, Maltese terrier, Maltese',
    154:
        'Pekinese, Pekingese, Peke',
    155:
        'Shih-Tzu',
    156:
        'Blenheim spaniel',
    157:
        'papillon',
    158:
        'toy terrier',
    159:
        'Rhodesian ridgeback',
    160:
        'Afghan hound, Afghan',
    161:
        'basset, basset hound',
    162:
        'beagle',
    163:
        'bloodhound, sleuthhound',
    164:
        'bluetick',
    165:
        'black-and-tan coonhound',
    166:
        'Walker hound, Walker foxhound',
    167:
        'English foxhound',
    168:
        'redbone',
    169:
        'borzoi, Russian wolfhound',
    170:
        'Irish wolfhound',
    171:
        'Italian greyhound',
    172:
        'whippet',
    173:
        'Ibizan hound, Ibizan Podenco',
    174:
        'Norwegian elkhound, elkhound',
    175:
        'otterhound, otter hound',
    176:
        'Saluki, gazelle hound',
    177:
        'Scottish deerhound, deerhound',
    178:
        'Weimaraner',
    179:
        'Staffordshire bullterrier, Staffordshire bull terrier',
    180:
        'American Staffordshire terrier, Staffordshire terrier, American pit bull terrier, pit bull terrier',
    181:
        'Bedlington terrier',
    182:
        'Border terrier',
    183:
        'Kerry blue terrier',
    184:
        'Irish terrier',
    185:
        'Norfolk terrier',
    186:
        'Norwich terrier',
    187:
        'Yorkshire terrier',
    188:
        'wire-haired fox terrier',
    189:
        'Lakeland terrier',
    190:
        'Sealyham terrier, Sealyham',
    191:
        'Airedale, Airedale terrier',
    192:
        'cairn, cairn terrier',
    193:
        'Australian terrier',
    194:
        'Dandie Dinmont, Dandie Dinmont terrier',
    195:
        'Boston bull, Boston terrier',
    196:
        'miniature schnauzer',
    197:
        'giant schnauzer',
    198:
        'standard schnauzer',
    199:
        'Scotch terrier, Scottish terrier, Scottie',
    200:
        'Tibetan terrier, chrysanthemum dog',
    201:
        'silky terrier, Sydney silky',
    202:
        'soft-coated wheaten terrier',
    203:
        'West Highland white terrier',
    204:
        'Lhasa, Lhasa apso',
    205:
        'flat-coated retriever',
    206:
        'curly-coated retriever',
    207:
        'golden retriever',
    208:
        'Labrador retriever',
    209:
        'Chesapeake Bay retriever',
    210:
        'German short-haired pointer',
    211:
        'vizsla, Hungarian pointer',
    212:
        'English setter',
    213:
        'Irish setter, red setter',
    214:
        'Gordon setter',
    215:
        'Brittany spaniel',
    216:
        'clumber, clumber spaniel',
    217:
        'English springer, English springer spaniel',
    218:
        'Welsh springer spaniel',
    219:
        'cocker spaniel, English cocker spaniel, cocker',
    220:
        'Sussex spaniel',
    221:
        'Irish water spaniel',
    222:
        'kuvasz',
    223:
        'schipperke',
    224:
        'groenendael',
    225:
        'malinois',
    226:
        'briard',
    227:
        'kelpie',
    228:
        'komondor',
    229:
        'Old English sheepdog, bobtail',
    230:
        'Shetland sheepdog, Shetland sheep dog, Shetland',
    231:
        'collie',
    232:
        'Border collie',
    233:
        'Bouvier des Flandres, Bouviers des Flandres',
    234:
        'Rottweiler',
    235:
        'German shepherd, German shepherd dog, German police dog, alsatian',
    236:
        'Doberman, Doberman pinscher',
    237:
        'miniature pinscher',
    238:
        'Greater Swiss Mountain dog',
    239:
        'Bernese mountain dog',
    240:
        'Appenzeller',
    241:
        'EntleBucher',
    242:
        'boxer',
    243:
        'bull mastiff',
    244:
        'Tibetan mastiff',
    245:
        'French bulldog',
    246:
        'Great Dane',
    247:
        'Saint Bernard, St Bernard',
    248:
        'Eskimo dog, husky',
    249:
        'malamute, malemute, Alaskan malamute',
    250:
        'Siberian husky',
    251:
        'dalmatian, coach dog, carriage dog',
    252:
        'affenpinscher, monkey pinscher, monkey dog',
    253:
        'basenji',
    254:
        'pug, pug-dog',
    255:
        'Leonberg',
    256:
        'Newfoundland, Newfoundland dog',
    257:
        'Great Pyrenees',
    258:
        'Samoyed, Samoyede',
    259:
        'Pomeranian',
    260:
        'chow, chow chow',
    261:
        'keeshond',
    262:
        'Brabancon griffon',
    263:
        'Pembroke, Pembroke Welsh corgi',
    264:
        'Cardigan, Cardigan Welsh corgi',
    265:
        'toy poodle',
    266:
        'miniature poodle',
    267:
        'standard poodle',
    268:
        'Mexican hairless',
    269:
        'timber wolf, grey wolf, gray wolf, Canis lupus',
    270:
        'white wolf, Arctic wolf, Canis lupus tundrarum',
    271:
        'red wolf, maned wolf, Canis rufus, Canis niger',
    272:
        'coyote, prairie wolf, brush wolf, Canis latrans',
    273:
        'dingo, warrigal, warragal, Canis dingo',
    274:
        'dhole, Cuon alpinus',
    275:
        'African hunting dog, hyena dog, Cape hunting dog, Lycaon pictus',
    276:
        'hyena, hyaena',
    277:
        'red fox, Vulpes vulpes',
    278:
        'kit fox, Vulpes macrotis',
    279:
        'Arctic fox, white fox, Alopex lagopus',
    280:
        'grey fox, gray fox, Urocyon cinereoargenteus',
    281:
        'tabby, tabby cat',
    282:
        'tiger cat',
    283:
        'Persian cat',
    284:
        'Siamese cat, Siamese',
    285:
        'Egyptian cat',
    286:
        'cougar, puma, catamount, mountain lion, painter, panther, Felis concolor',
    287:
        'lynx, catamount',
    288:
        'leopard, Panthera pardus',
    289:
        'snow leopard, ounce, Panthera uncia',
    290:
        'jaguar, panther, Panthera onca, Felis onca',
    291:
        'lion, king of beasts, Panthera leo',
    292:
        'tiger, Panthera tigris',
    293:
        'cheetah, chetah, Acinonyx jubatus',
    294:
        'brown bear, bruin, Ursus arctos',
    295:
        'American black bear, black bear, Ursus americanus, Euarctos americanus',
    296:
        'ice bear, polar bear, Ursus Maritimus, Thalarctos maritimus',
    297:
        'sloth bear, Melursus ursinus, Ursus ursinus',
    298:
        'mongoose',
    299:
        'meerkat, mierkat',
    300:
        'tiger beetle',
    301:
        'ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle',
    302:
        'ground beetle, carabid beetle',
    303:
        'long-horned beetle, longicorn, longicorn beetle',
    304:
        'leaf beetle, chrysomelid',
    305:
        'dung beetle',
    306:
        'rhinoceros beetle',
    307:
        'weevil',
    308:
        'fly',
    309:
        'bee',
    310:
        'ant, emmet, pismire',
    311:
        'grasshopper, hopper',
    312:
        'cricket',
    313:
        'walking stick, walkingstick, stick insect',
    314:
        'cockroach, roach',
    315:
        'mantis, mantid',
    316:
        'cicada, cicala',
    317:
        'leafhopper',
    318:
        'lacewing, lacewing fly',
    319:
        "dragonfly, darning needle, devil's darning needle, sewing needle, snake feeder, snake doctor, mosquito hawk, skeeter hawk",
    320:
        'damselfly',
    321:
        'admiral',
    322:
        'ringlet, ringlet butterfly',
    323:
        'monarch, monarch butterfly, milkweed butterfly, Danaus plexippus',
    324:
        'cabbage butterfly',
    325:
        'sulphur butterfly, sulfur butterfly',
    326:
        'lycaenid, lycaenid butterfly',
    327:
        'starfish, sea star',
    328:
        'sea urchin',
    329:
        'sea cucumber, holothurian',
    330:
        'wood rabbit, cottontail, cottontail rabbit',
    331:
        'hare',
    332:
        'Angora, Angora rabbit',
    333:
        'hamster',
    334:
        'porcupine, hedgehog',
    335:
        'fox squirrel, eastern fox squirrel, Sciurus niger',
    336:
        'marmot',
    337:
        'beaver',
    338:
        'guinea pig, Cavia cobaya',
    339:
        'sorrel',
    340:
        'zebra',
    341:
        'hog, pig, grunter, squealer, Sus scrofa',
    342:
        'wild boar, boar, Sus scrofa',
    343:
        'warthog',
    344:
        'hippopotamus, hippo, river horse, Hippopotamus amphibius',
    345:
        'ox',
    346:
        'water buffalo, water ox, Asiatic buffalo, Bubalus bubalis',
    347:
        'bison',
    348:
        'ram, tup',
    349:
        'bighorn, bighorn sheep, cimarron, Rocky Mountain bighorn, Rocky Mountain sheep, Ovis canadensis',
    350:
        'ibex, Capra ibex',
    351:
        'hartebeest',
    352:
        'impala, Aepyceros melampus',
    353:
        'gazelle',
    354:
        'Arabian camel, dromedary, Camelus dromedarius',
    355:
        'llama',
    356:
        'weasel',
    357:
        'mink',
    358:
        'polecat, fitch, foulmart, foumart, Mustela putorius',
    359:
        'black-footed ferret, ferret, Mustela nigripes',
    360:
        'otter',
    361:
        'skunk, polecat, wood pussy',
    362:
        'badger',
    363:
        'armadillo',
    364:
        'three-toed sloth, ai, Bradypus tridactylus',
    365:
        'orangutan, orang, orangutang, Pongo pygmaeus',
    366:
        'gorilla, Gorilla gorilla',
    367:
        'chimpanzee, chimp, Pan troglodytes',
    368:
        'gibbon, Hylobates lar',
    369:
        'siamang, Hylobates syndactylus, Symphalangus syndactylus',
    370:
        'guenon, guenon monkey',
    371:
        'patas, hussar monkey, Erythrocebus patas',
    372:
        'baboon',
    373:
        'macaque',
    374:
        'langur',
    375:
        'colobus, colobus monkey',
    376:
        'proboscis monkey, Nasalis larvatus',
    377:
        'marmoset',
    378:
        'capuchin, ringtail, Cebus capucinus',
    379:
        'howler monkey, howler',
    380:
        'titi, titi monkey',
    381:
        'spider monkey, Ateles geoffroyi',
    382:
        'squirrel monkey, Saimiri sciureus',
    383:
        'Madagascar cat, ring-tailed lemur, Lemur catta',
    384:
        'indri, indris, Indri indri, Indri brevicaudatus',
    385:
        'Indian elephant, Elephas maximus',
    386:
        'African elephant, Loxodonta africana',
    387:
        'lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens',
    388:
        'giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca',
    389:
        'barracouta, snoek',
    390:
        'eel',
    391:
        'coho, cohoe, coho salmon, blue jack, silver salmon, Oncorhynchus kisutch',
    392:
        'rock beauty, Holocanthus tricolor',
    393:
        'anemone fish',
    394:
        'sturgeon',
    395:
        'gar, garfish, garpike, billfish, Lepisosteus osseus',
    396:
        'lionfish',
    397:
        'puffer, pufferfish, blowfish, globefish',
    398:
        'abacus',
    399:
        'abaya',
    400:
        "academic gown, academic robe, judge's robe",
    401:
        'accordion, piano accordion, squeeze box',
    402:
        'acoustic guitar',
    403:
        'aircraft carrier, carrier, flattop, attack aircraft carrier',
    404:
        'airliner',
    405:
        'airship, dirigible',
    406:
        'altar',
    407:
        'ambulance',
    408:
        'amphibian, amphibious vehicle',
    409:
        'analog clock',
    410:
        'apiary, bee house',
    411:
        'apron',
    412:
        'ashcan, trash can, garbage can, wastebin, ash bin, ash-bin, ashbin, dustbin, trash barrel, trash bin',
    413:
        'assault rifle, assault gun',
    414:
        'backpack, back pack, knapsack, packsack, rucksack, haversack',
    415:
        'bakery, bakeshop, bakehouse',
    416:
        'balance beam, beam',
    417:
        'balloon',
    418:
        'ballpoint, ballpoint pen, ballpen, Biro',
    419:
        'Band Aid',
    420:
        'banjo',
    421:
        'bannister, banister, balustrade, balusters, handrail',
    422:
        'barbell',
    423:
        'barber chair',
    424:
        'barbershop',
    425:
        'barn',
    426:
        'barometer',
    427:
        'barrel, cask',
    428:
        'barrow, garden cart, lawn cart, wheelbarrow',
    429:
        'baseball',
    430:
        'basketball',
    431:
        'bassinet',
    432:
        'bassoon',
    433:
        'bathing cap, swimming cap',
    434:
        'bath towel',
    435:
        'bathtub, bathing tub, bath, tub',
    436:
        'beach wagon, station wagon, wagon, estate car, beach waggon, station waggon, waggon',
    437:
        'beacon, lighthouse, beacon light, pharos',
    438:
        'beaker',
    439:
        'bearskin, busby, shako',
    440:
        'beer bottle',
    441:
        'beer glass',
    442:
        'bell cote, bell cot',
    443:
        'bib',
    444:
        'bicycle-built-for-two, tandem bicycle, tandem',
    445:
        'bikini, two-piece',
    446:
        'binder, ring-binder',
    447:
        'binoculars, field glasses, opera glasses',
    448:
        'birdhouse',
    449:
        'boathouse',
    450:
        'bobsled, bobsleigh, bob',
    451:
        'bolo tie, bolo, bola tie, bola',
    452:
        'bonnet, poke bonnet',
    453:
        'bookcase',
    454:
        'bookshop, bookstore, bookstall',
    455:
        'bottlecap',
    456:
        'bow',
    457:
        'bow tie, bow-tie, bowtie',
    458:
        'brass, memorial tablet, plaque',
    459:
        'brassiere, bra, bandeau',
    460:
        'breakwater, groin, groyne, mole, bulwark, seawall, jetty',
    461:
        'breastplate, aegis, egis',
    462:
        'broom',
    463:
        'bucket, pail',
    464:
        'buckle',
    465:
        'bulletproof vest',
    466:
        'bullet train, bullet',
    467:
        'butcher shop, meat market',
    468:
        'cab, hack, taxi, taxicab',
    469:
        'caldron, cauldron',
    470:
        'candle, taper, wax light',
    471:
        'cannon',
    472:
        'canoe',
    473:
        'can opener, tin opener',
    474:
        'cardigan',
    475:
        'car mirror',
    476:
        'carousel, carrousel, merry-go-round, roundabout, whirligig',
    477:
        "carpenter's kit, tool kit",
    478:
        'carton',
    479:
        'car wheel',
    480:
        'cash machine, cash dispenser, automated teller machine, automatic teller machine, automated teller, automatic teller, ATM',
    481:
        'cassette',
    482:
        'cassette player',
    483:
        'castle',
    484:
        'catamaran',
    485:
        'CD player',
    486:
        'cello, violoncello',
    487:
        'cellular telephone, cellular phone, cellphone, cell, mobile phone',
    488:
        'chain',
    489:
        'chainlink fence',
    490:
        'chain mail, ring mail, mail, chain armor, chain armour, ring armor, ring armour',
    491:
        'chain saw, chainsaw',
    492:
        'chest',
    493:
        'chiffonier, commode',
    494:
        'chime, bell, gong',
    495:
        'china cabinet, china closet',
    496:
        'Christmas stocking',
    497:
        'church, church building',
    498:
        'cinema, movie theater, movie theatre, movie house, picture palace',
    499:
        'cleaver, meat cleaver, chopper',
    500:
        'cliff dwelling',
    501:
        'cloak',
    502:
        'clog, geta, patten, sabot',
    503:
        'cocktail shaker',
    504:
        'coffee mug',
    505:
        'coffeepot',
    506:
        'coil, spiral, volute, whorl, helix',
    507:
        'combination lock',
    508:
        'computer keyboard, keypad',
    509:
        'confectionery, confectionary, candy store',
    510:
        'container ship, containership, container vessel',
    511:
        'convertible',
    512:
        'corkscrew, bottle screw',
    513:
        'cornet, horn, trumpet, trump',
    514:
        'cowboy boot',
    515:
        'cowboy hat, ten-gallon hat',
    516:
        'cradle',
    517:
        'crane',
    518:
        'crash helmet',
    519:
        'crate',
    520:
        'crib, cot',
    521:
        'Crock Pot',
    522:
        'croquet ball',
    523:
        'crutch',
    524:
        'cuirass',
    525:
        'dam, dike, dyke',
    526:
        'desk',
    527:
        'desktop computer',
    528:
        'dial telephone, dial phone',
    529:
        'diaper, nappy, napkin',
    530:
        'digital clock',
    531:
        'digital watch',
    532:
        'dining table, board',
    533:
        'dishrag, dishcloth',
    534:
        'dishwasher, dish washer, dishwashing machine',
    535:
        'disk brake, disc brake',
    536:
        'dock, dockage, docking facility',
    537:
        'dogsled, dog sled, dog sleigh',
    538:
        'dome',
    539:
        'doormat, welcome mat',
    540:
        'drilling platform, offshore rig',
    541:
        'drum, membranophone, tympan',
    542:
        'drumstick',
    543:
        'dumbbell',
    544:
        'Dutch oven',
    545:
        'electric fan, blower',
    546:
        'electric guitar',
    547:
        'electric locomotive',
    548:
        'entertainment center',
    549:
        'envelope',
    550:
        'espresso maker',
    551:
        'face powder',
    552:
        'feather boa, boa',
    553:
        'file, file cabinet, filing cabinet',
    554:
        'fireboat',
    555:
        'fire engine, fire truck',
    556:
        'fire screen, fireguard',
    557:
        'flagpole, flagstaff',
    558:
        'flute, transverse flute',
    559:
        'folding chair',
    560:
        'football helmet',
    561:
        'forklift',
    562:
        'fountain',
    563:
        'fountain pen',
    564:
        'four-poster',
    565:
        'freight car',
    566:
        'French horn, horn',
    567:
        'frying pan, frypan, skillet',
    568:
        'fur coat',
    569:
        'garbage truck, dustcart',
    570:
        'gasmask, respirator, gas helmet',
    571:
        'gas pump, gasoline pump, petrol pump, island dispenser',
    572:
        'goblet',
    573:
        'go-kart',
    574:
        'golf ball',
    575:
        'golfcart, golf cart',
    576:
        'gondola',
    577:
        'gong, tam-tam',
    578:
        'gown',
    579:
        'grand piano, grand',
    580:
        'greenhouse, nursery, glasshouse',
    581:
        'grille, radiator grille',
    582:
        'grocery store, grocery, food market, market',
    583:
        'guillotine',
    584:
        'hair slide',
    585:
        'hair spray',
    586:
        'half track',
    587:
        'hammer',
    588:
        'hamper',
    589:
        'hand blower, blow dryer, blow drier, hair dryer, hair drier',
    590:
        'hand-held computer, hand-held microcomputer',
    591:
        'handkerchief, hankie, hanky, hankey',
    592:
        'hard disc, hard disk, fixed disk',
    593:
        'harmonica, mouth organ, harp, mouth harp',
    594:
        'harp',
    595:
        'harvester, reaper',
    596:
        'hatchet',
    597:
        'holster',
    598:
        'home theater, home theatre',
    599:
        'honeycomb',
    600:
        'hook, claw',
    601:
        'hoopskirt, crinoline',
    602:
        'horizontal bar, high bar',
    603:
        'horse cart, horse-cart',
    604:
        'hourglass',
    605:
        'iPod',
    606:
        'iron, smoothing iron',
    607:
        "jack-o'-lantern",
    608:
        'jean, blue jean, denim',
    609:
        'jeep, landrover',
    610:
        'jersey, T-shirt, tee shirt',
    611:
        'jigsaw puzzle',
    612:
        'jinrikisha, ricksha, rickshaw',
    613:
        'joystick',
    614:
        'kimono',
    615:
        'knee pad',
    616:
        'knot',
    617:
        'lab coat, laboratory coat',
    618:
        'ladle',
    619:
        'lampshade, lamp shade',
    620:
        'laptop, laptop computer',
    621:
        'lawn mower, mower',
    622:
        'lens cap, lens cover',
    623:
        'letter opener, paper knife, paperknife',
    624:
        'library',
    625:
        'lifeboat',
    626:
        'lighter, light, igniter, ignitor',
    627:
        'limousine, limo',
    628:
        'liner, ocean liner',
    629:
        'lipstick, lip rouge',
    630:
        'Loafer',
    631:
        'lotion',
    632:
        'loudspeaker, speaker, speaker unit, loudspeaker system, speaker system',
    633:
        "loupe, jeweler's loupe",
    634:
        'lumbermill, sawmill',
    635:
        'magnetic compass',
    636:
        'mailbag, postbag',
    637:
        'mailbox, letter box',
    638:
        'maillot',
    639:
        'maillot, tank suit',
    640:
        'manhole cover',
    641:
        'maraca',
    642:
        'marimba, xylophone',
    643:
        'mask',
    644:
        'matchstick',
    645:
        'maypole',
    646:
        'maze, labyrinth',
    647:
        'measuring cup',
    648:
        'medicine chest, medicine cabinet',
    649:
        'megalith, megalithic structure',
    650:
        'microphone, mike',
    651:
        'microwave, microwave oven',
    652:
        'military uniform',
    653:
        'milk can',
    654:
        'minibus',
    655:
        'miniskirt, mini',
    656:
        'minivan',
    657:
        'missile',
    658:
        'mitten',
    659:
        'mixing bowl',
    660:
        'mobile home, manufactured home',
    661:
        'Model T',
    662:
        'modem',
    663:
        'monastery',
    664:
        'monitor',
    665:
        'moped',
    666:
        'mortar',
    667:
        'mortarboard',
    668:
        'mosque',
    669:
        'mosquito net',
    670:
        'motor scooter, scooter',
    671:
        'mountain bike, all-terrain bike, off-roader',
    672:
        'mountain tent',
    673:
        'mouse, computer mouse',
    674:
        'mousetrap',
    675:
        'moving van',
    676:
        'muzzle',
    677:
        'nail',
    678:
        'neck brace',
    679:
        'necklace',
    680:
        'nipple',
    681:
        'notebook, notebook computer',
    682:
        'obelisk',
    683:
        'oboe, hautboy, hautbois',
    684:
        'ocarina, sweet potato',
    685:
        'odometer, hodometer, mileometer, milometer',
    686:
        'oil filter',
    687:
        'organ, pipe organ',
    688:
        'oscilloscope, scope, cathode-ray oscilloscope, CRO',
    689:
        'overskirt',
    690:
        'oxcart',
    691:
        'oxygen mask',
    692:
        'packet',
    693:
        'paddle, boat paddle',
    694:
        'paddlewheel, paddle wheel',
    695:
        'padlock',
    696:
        'paintbrush',
    697:
        "pajama, pyjama, pj's, jammies",
    698:
        'palace',
    699:
        'panpipe, pandean pipe, syrinx',
    700:
        'paper towel',
    701:
        'parachute, chute',
    702:
        'parallel bars, bars',
    703:
        'park bench',
    704:
        'parking meter',
    705:
        'passenger car, coach, carriage',
    706:
        'patio, terrace',
    707:
        'pay-phone, pay-station',
    708:
        'pedestal, plinth, footstall',
    709:
        'pencil box, pencil case',
    710:
        'pencil sharpener',
    711:
        'perfume, essence',
    712:
        'Petri dish',
    713:
        'photocopier',
    714:
        'pick, plectrum, plectron',
    715:
        'pickelhaube',
    716:
        'picket fence, paling',
    717:
        'pickup, pickup truck',
    718:
        'pier',
    719:
        'piggy bank, penny bank',
    720:
        'pill bottle',
    721:
        'pillow',
    722:
        'ping-pong ball',
    723:
        'pinwheel',
    724:
        'pirate, pirate ship',
    725:
        'pitcher, ewer',
    726:
        "plane, carpenter's plane, woodworking plane",
    727:
        'planetarium',
    728:
        'plastic bag',
    729:
        'plate rack',
    730:
        'plow, plough',
    731:
        "plunger, plumber's helper",
    732:
        'Polaroid camera, Polaroid Land camera',
    733:
        'pole',
    734:
        'police van, police wagon, paddy wagon, patrol wagon, wagon, black Maria',
    735:
        'poncho',
    736:
        'pool table, billiard table, snooker table',
    737:
        'pop bottle, soda bottle',
    738:
        'pot, flowerpot',
    739:
        "potter's wheel",
    740:
        'power drill',
    741:
        'prayer rug, prayer mat',
    742:
        'printer',
    743:
        'prison, prison house',
    744:
        'projectile, missile',
    745:
        'projector',
    746:
        'puck, hockey puck',
    747:
        'punching bag, punch bag, punching ball, punchball',
    748:
        'purse',
    749:
        'quill, quill pen',
    750:
        'quilt, comforter, comfort, puff',
    751:
        'racer, race car, racing car',
    752:
        'racket, racquet',
    753:
        'radiator',
    754:
        'radio, wireless',
    755:
        'radio telescope, radio reflector',
    756:
        'rain barrel',
    757:
        'recreational vehicle, RV, R.V.',
    758:
        'reel',
    759:
        'reflex camera',
    760:
        'refrigerator, icebox',
    761:
        'remote control, remote',
    762:
        'restaurant, eating house, eating place, eatery',
    763:
        'revolver, six-gun, six-shooter',
    764:
        'rifle',
    765:
        'rocking chair, rocker',
    766:
        'rotisserie',
    767:
        'rubber eraser, rubber, pencil eraser',
    768:
        'rugby ball',
    769:
        'rule, ruler',
    770:
        'running shoe',
    771:
        'safe',
    772:
        'safety pin',
    773:
        'saltshaker, salt shaker',
    774:
        'sandal',
    775:
        'sarong',
    776:
        'sax, saxophone',
    777:
        'scabbard',
    778:
        'scale, weighing machine',
    779:
        'school bus',
    780:
        'schooner',
    781:
        'scoreboard',
    782:
        'screen, CRT screen',
    783:
        'screw',
    784:
        'screwdriver',
    785:
        'seat belt, seatbelt',
    786:
        'sewing machine',
    787:
        'shield, buckler',
    788:
        'shoe shop, shoe-shop, shoe store',
    789:
        'shoji',
    790:
        'shopping basket',
    791:
        'shopping cart',
    792:
        'shovel',
    793:
        'shower cap',
    794:
        'shower curtain',
    795:
        'ski',
    796:
        'ski mask',
    797:
        'sleeping bag',
    798:
        'slide rule, slipstick',
    799:
        'sliding door',
    800:
        'slot, one-armed bandit',
    801:
        'snorkel',
    802:
        'snowmobile',
    803:
        'snowplow, snowplough',
    804:
        'soap dispenser',
    805:
        'soccer ball',
    806:
        'sock',
    807:
        'solar dish, solar collector, solar furnace',
    808:
        'sombrero',
    809:
        'soup bowl',
    810:
        'space bar',
    811:
        'space heater',
    812:
        'space shuttle',
    813:
        'spatula',
    814:
        'speedboat',
    815:
        "spider web, spider's web",
    816:
        'spindle',
    817:
        'sports car, sport car',
    818:
        'spotlight, spot',
    819:
        'stage',
    820:
        'steam locomotive',
    821:
        'steel arch bridge',
    822:
        'steel drum',
    823:
        'stethoscope',
    824:
        'stole',
    825:
        'stone wall',
    826:
        'stopwatch, stop watch',
    827:
        'stove',
    828:
        'strainer',
    829:
        'streetcar, tram, tramcar, trolley, trolley car',
    830:
        'stretcher',
    831:
        'studio couch, day bed',
    832:
        'stupa, tope',
    833:
        'submarine, pigboat, sub, U-boat',
    834:
        'suit, suit of clothes',
    835:
        'sundial',
    836:
        'sunglass',
    837:
        'sunglasses, dark glasses, shades',
    838:
        'sunscreen, sunblock, sun blocker',
    839:
        'suspension bridge',
    840:
        'swab, swob, mop',
    841:
        'sweatshirt',
    842:
        'swimming trunks, bathing trunks',
    843:
        'swing',
    844:
        'switch, electric switch, electrical switch',
    845:
        'syringe',
    846:
        'table lamp',
    847:
        'tank, army tank, armored combat vehicle, armoured combat vehicle',
    848:
        'tape player',
    849:
        'teapot',
    850:
        'teddy, teddy bear',
    851:
        'television, television system',
    852:
        'tennis ball',
    853:
        'thatch, thatched roof',
    854:
        'theater curtain, theatre curtain',
    855:
        'thimble',
    856:
        'thresher, thrasher, threshing machine',
    857:
        'throne',
    858:
        'tile roof',
    859:
        'toaster',
    860:
        'tobacco shop, tobacconist shop, tobacconist',
    861:
        'toilet seat',
    862:
        'torch',
    863:
        'totem pole',
    864:
        'tow truck, tow car, wrecker',
    865:
        'toyshop',
    866:
        'tractor',
    867:
        'trailer truck, tractor trailer, trucking rig, rig, articulated lorry, semi',
    868:
        'tray',
    869:
        'trench coat',
    870:
        'tricycle, trike, velocipede',
    871:
        'trimaran',
    872:
        'tripod',
    873:
        'triumphal arch',
    874:
        'trolleybus, trolley coach, trackless trolley',
    875:
        'trombone',
    876:
        'tub, vat',
    877:
        'turnstile',
    878:
        'typewriter keyboard',
    879:
        'umbrella',
    880:
        'unicycle, monocycle',
    881:
        'upright, upright piano',
    882:
        'vacuum, vacuum cleaner',
    883:
        'vase',
    884:
        'vault',
    885:
        'velvet',
    886:
        'vending machine',
    887:
        'vestment',
    888:
        'viaduct',
    889:
        'violin, fiddle',
    890:
        'volleyball',
    891:
        'waffle iron',
    892:
        'wall clock',
    893:
        'wallet, billfold, notecase, pocketbook',
    894:
        'wardrobe, closet, press',
    895:
        'warplane, military plane',
    896:
        'washbasin, handbasin, washbowl, lavabo, wash-hand basin',
    897:
        'washer, automatic washer, washing machine',
    898:
        'water bottle',
    899:
        'water jug',
    900:
        'water tower',
    901:
        'whiskey jug',
    902:
        'whistle',
    903:
        'wig',
    904:
        'window screen',
    905:
        'window shade',
    906:
        'Windsor tie',
    907:
        'wine bottle',
    908:
        'wing',
    909:
        'wok',
    910:
        'wooden spoon',
    911:
        'wool, woolen, woollen',
    912:
        'worm fence, snake fence, snake-rail fence, Virginia fence',
    913:
        'wreck',
    914:
        'yawl',
    915:
        'yurt',
    916:
        'web site, website, internet site, site',
    917:
        'comic book',
    918:
        'crossword puzzle, crossword',
    919:
        'street sign',
    920:
        'traffic light, traffic signal, stoplight',
    921:
        'book jacket, dust cover, dust jacket, dust wrapper',
    922:
        'menu',
    923:
        'plate',
    924:
        'guacamole',
    925:
        'consomme',
    926:
        'hot pot, hotpot',
    927:
        'trifle',
    928:
        'ice cream, icecream',
    929:
        'ice lolly, lolly, lollipop, popsicle',
    930:
        'French loaf',
    931:
        'bagel, beigel',
    932:
        'pretzel',
    933:
        'cheeseburger',
    934:
        'hotdog, hot dog, red hot',
    935:
        'mashed potato',
    936:
        'head cabbage',
    937:
        'broccoli',
    938:
        'cauliflower',
    939:
        'zucchini, courgette',
    940:
        'spaghetti squash',
    941:
        'acorn squash',
    942:
        'butternut squash',
    943:
        'cucumber, cuke',
    944:
        'artichoke, globe artichoke',
    945:
        'bell pepper',
    946:
        'cardoon',
    947:
        'mushroom',
    948:
        'Granny Smith',
    949:
        'strawberry',
    950:
        'orange',
    951:
        'lemon',
    952:
        'fig',
    953:
        'pineapple, ananas',
    954:
        'banana',
    955:
        'jackfruit, jak, jack',
    956:
        'custard apple',
    957:
        'pomegranate',
    958:
        'hay',
    959:
        'carbonara',
    960:
        'chocolate sauce, chocolate syrup',
    961:
        'dough',
    962:
        'meat loaf, meatloaf',
    963:
        'pizza, pizza pie',
    964:
        'potpie',
    965:
        'burrito',
    966:
        'red wine',
    967:
        'espresso',
    968:
        'cup',
    969:
        'eggnog',
    970:
        'alp',
    971:
        'bubble',
    972:
        'cliff, drop, drop-off',
    973:
        'coral reef',
    974:
        'geyser',
    975:
        'lakeside, lakeshore',
    976:
        'promontory, headland, head, foreland',
    977:
        'sandbar, sand bar',
    978:
        'seashore, coast, seacoast, sea-coast',
    979:
        'valley, vale',
    980:
        'volcano',
    981:
        'ballplayer, baseball player',
    982:
        'groom, bridegroom',
    983:
        'scuba diver',
    984:
        'rapeseed',
    985:
        'daisy',
    986:
        "yellow lady's slipper, yellow lady-slipper, Cypripedium calceolus, Cypripedium parviflorum",
    987:
        'corn',
    988:
        'acorn',
    989:
        'hip, rose hip, rosehip',
    990:
        'buckeye, horse chestnut, conker',
    991:
        'coral fungus',
    992:
        'agaric',
    993:
        'gyromitra',
    994:
        'stinkhorn, carrion fungus',
    995:
        'earthstar',
    996:
        'hen-of-the-woods, hen of the woods, Polyporus frondosus, Grifola frondosa',
    997:
        'bolete',
    998:
        'ear, spike, capitulum',
    999:
        'toilet tissue, toilet paper, bathroom tissue'
}



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Provides utilities to preprocess images.

Training images are sampled using the provided bounding boxes, and subsequently
cropped to the sampled bounding box. Images are additionally flipped randomly,
then resized to the target output size (without aspect-ratio preservation).

Images used during evaluation are resized (with aspect-ratio preservation) and
centrally cropped.

All images undergo mean color subtraction.

Note that these steps are colloquially referred to as "ResNet preprocessing,"
and they differ from "VGG preprocessing," which does not use bounding boxes
and instead does an aspect-preserving resize followed by random crop during
training. (These both differ from "Inception preprocessing," which introduces
color distortion steps.)

Brainchip changes:
        - Original file: https://github.com/tensorflow/models/blob/master/official/vision/image_classification/imagenet_preprocessing.py
        - Update preprocess_image() return value for Akida compatibility
        - Change cycle_length 'parallelization' parameter
        - Add _distort_color, _apply_with_random_selector, and _random_distort_color methods
          from https://github.com/tensorflow/models/blob/master/research/slim/preprocessing/inception_preprocessing.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from absl import logging
import tensorflow as tf
import sys
import numpy as np

from tensorflow.python.ops import control_flow_ops
from .imagenet_labels2names import imagenet_labels

NUM_CHANNELS = 3
NUM_CLASSES = 1001

NUM_IMAGES = {
    'train': 1281167,
    'validation': 50000,
}

_NUM_TRAIN_FILES = 1024
_SHUFFLE_BUFFER = 10000

_R_MEAN = 123.68
_G_MEAN = 116.78
_B_MEAN = 103.94
CHANNEL_MEANS = [_R_MEAN, _G_MEAN, _B_MEAN]

# The lower bound for the smallest side of the image for aspect-preserving
# resizing. For example, if an image is 500 x 1000, it will be resized to
# _RESIZE_MIN x (_RESIZE_MIN * 2).
_RESIZE_MIN = 256


def process_record_dataset(dataset,
                           is_training,
                           batch_size,
                           im_size,
                           shuffle_buffer,
                           parse_record_fn,
                           num_epochs=1,
                           dtype=tf.float32,
                           datasets_num_private_threads=None,
                           drop_remainder=False,
                           tf_data_experimental_slack=False):
    """Given a Dataset with raw records, return an iterator over the records.

  Args:
    dataset: A Dataset representing raw records
    is_training: A boolean denoting whether the input is for training.
    batch_size: The number of samples per batch.
    shuffle_buffer: The buffer size to use when shuffling records. A larger
      value results in better randomness, but smaller values reduce startup
      time and use less memory.
    parse_record_fn: A function that takes a raw record and returns the
      corresponding (image, label) pair.
    num_epochs: The number of epochs to repeat the dataset.
    dtype: Data type to use for images/features.
    datasets_num_private_threads: Number of threads for a private
      threadpool created for all datasets computation.
    drop_remainder: A boolean indicates whether to drop the remainder of the
      batches. If True, the batch dimension will be static.
    tf_data_experimental_slack: Whether to enable tf.data's
      `experimental_slack` option.

  Returns:
    Dataset of (image, label) pairs ready for iteration.
  """
    # Defines a specific size thread pool for tf.data operations.
    if datasets_num_private_threads:
        options = tf.data.Options()
        options.experimental_threading.private_threadpool_size = (
            datasets_num_private_threads)
        dataset = dataset.with_options(options)
        logging.info('datasets_num_private_threads: %s',
                     datasets_num_private_threads)

    if is_training:
        # Shuffles records before repeating to respect epoch boundaries.
        dataset = dataset.shuffle(buffer_size=shuffle_buffer)
        # Repeats the dataset for the number of epochs to train.
        dataset = dataset.repeat()

    # Parses the raw records into images and labels.
    dataset = dataset.map(
        lambda value: parse_record_fn(value, im_size, is_training, dtype),
        num_parallel_calls=tf.data.experimental.AUTOTUNE)
    dataset = dataset.batch(batch_size, drop_remainder=drop_remainder)

    # Operations between the final prefetch and the get_next call to the iterator
    # will happen synchronously during run time. We prefetch here again to
    # background all of the above processing work and keep it out of the
    # critical training path. Setting buffer_size to tf.data.experimental.AUTOTUNE
    # allows DistributionStrategies to adjust how many batches to fetch based
    # on how many devices are present.
    dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)

    options = tf.data.Options()
    options.experimental_slack = tf_data_experimental_slack
    dataset = dataset.with_options(options)

    return dataset


def get_filenames(is_training, data_dir):
    """Return filenames for dataset."""
    if is_training:
        return [
            os.path.join(data_dir, 'train-%05d-of-01024' % i)
            for i in range(_NUM_TRAIN_FILES)
        ]
    else:
        return [
            os.path.join(data_dir, 'validation-%05d-of-00128' % i)
            for i in range(128)
        ]


def _parse_example_proto(example_serialized):
    """Parses an Example proto containing a training example of an image.

  The output of the build_image_data.py image preprocessing script is a dataset
  containing serialized Example protocol buffers. Each Example proto contains
  the following fields (values are included as examples):

    image/height: 462
    image/width: 581
    image/colorspace: 'RGB'
    image/channels: 3
    image/class/label: 615
    image/class/synset: 'n03623198'
    image/class/text: 'knee pad'
    image/object/bbox/xmin: 0.1
    image/object/bbox/xmax: 0.9
    image/object/bbox/ymin: 0.2
    image/object/bbox/ymax: 0.6
    image/object/bbox/label: 615
    image/format: 'JPEG'
    image/filename: 'ILSVRC2012_val_00041207.JPEG'
    image/encoded: <JPEG encoded string>

  Args:
    example_serialized: scalar Tensor tf.string containing a serialized
      Example protocol buffer.

  Returns:
    image_buffer: Tensor tf.string containing the contents of a JPEG file.
    label: Tensor tf.int32 containing the label.
    bbox: 3-D float Tensor of bounding boxes arranged [1, num_boxes, coords]
      where each coordinate is [0, 1) and the coordinates are arranged as
      [ymin, xmin, ymax, xmax].
  """
    # Dense features in Example proto.
    feature_map = {
        'image/encoded':
            tf.io.FixedLenFeature([], dtype=tf.string, default_value=''),
        'image/class/label':
            tf.io.FixedLenFeature([], dtype=tf.int64, default_value=-1),
        'image/class/text':
            tf.io.FixedLenFeature([], dtype=tf.string, default_value=''),
    }
    sparse_float32 = tf.io.VarLenFeature(dtype=tf.float32)
    # Sparse features in Example proto.
    feature_map.update({
        k: sparse_float32 for k in [
            'image/object/bbox/xmin', 'image/object/bbox/ymin',
            'image/object/bbox/xmax', 'image/object/bbox/ymax'
        ]
    })

    features = tf.io.parse_single_example(serialized=example_serialized,
                                          features=feature_map)
    label = tf.cast(features['image/class/label'], dtype=tf.int32)

    xmin = tf.expand_dims(features['image/object/bbox/xmin'].values, 0)
    ymin = tf.expand_dims(features['image/object/bbox/ymin'].values, 0)
    xmax = tf.expand_dims(features['image/object/bbox/xmax'].values, 0)
    ymax = tf.expand_dims(features['image/object/bbox/ymax'].values, 0)

    # Note that we impose an ordering of (y, x) just to make life difficult.
    bbox = tf.concat([ymin, xmin, ymax, xmax], 0)

    # Force the variable number of bounding boxes into the shape
    # [1, num_boxes, coords].
    bbox = tf.expand_dims(bbox, 0)
    bbox = tf.transpose(a=bbox, perm=[0, 2, 1])

    return features['image/encoded'], label, bbox


def parse_record(raw_record, im_size, is_training, dtype):
    """Parses a record containing a training example of an image.

  The input record is parsed into a label and image, and the image is passed
  through preprocessing steps (cropping, flipping, and so on).

  Args:
    raw_record: scalar Tensor tf.string containing a serialized
      Example protocol buffer.
    is_training: A boolean denoting whether the input is for training.
    dtype: data type to use for images/features.

  Returns:
    Tuple with processed image tensor and one-hot-encoded label tensor.
  """
    image_buffer, label, bbox = _parse_example_proto(raw_record)

    image = preprocess_image(image_buffer=image_buffer,
                             bbox=bbox,
                             output_height=im_size,
                             output_width=im_size,
                             num_channels=NUM_CHANNELS,
                             is_training=is_training)
    image = tf.cast(image, dtype)

    # Subtract one so that labels are in [0, 1000), and cast to float32 for
    # Keras model.
    label = tf.cast(tf.cast(tf.reshape(label, shape=[1]), dtype=tf.int32) - 1,
                    dtype=tf.float32)
    return image, label


def input_fn(is_training,
             data_dir,
             batch_size,
             im_size,
             num_epochs=1,
             dtype=tf.float32,
             datasets_num_private_threads=None,
             parse_record_fn=parse_record,
             input_context=None,
             drop_remainder=False,
             tf_data_experimental_slack=False,
             training_dataset_cache=False):
    """Input function which provides batches for train or eval.

  Args:
    is_training: A boolean denoting whether the input is for training.
    data_dir: The directory containing the input data.
    batch_size: The number of samples per batch.
    num_epochs: The number of epochs to repeat the dataset.
    dtype: Data type to use for images/features
    datasets_num_private_threads: Number of private threads for tf.data.
    parse_record_fn: Function to use for parsing the records.
    input_context: A `tf.distribute.InputContext` object passed in by
      `tf.distribute.Strategy`.
    drop_remainder: A boolean indicates whether to drop the remainder of the
      batches. If True, the batch dimension will be static.
    tf_data_experimental_slack: Whether to enable tf.data's
      `experimental_slack` option.
    training_dataset_cache: Whether to cache the training dataset on workers.
       Typically used to improve training performance when training data is in
       remote storage and can fit into worker memory.

  Returns:
    A dataset that can be used for iteration.
  """
    filenames = get_filenames(is_training, data_dir)
    dataset = tf.data.Dataset.from_tensor_slices(filenames)

    if input_context:
        logging.info(
            'Sharding the dataset: input_pipeline_id=%d num_input_pipelines=%d',
            input_context.input_pipeline_id, input_context.num_input_pipelines)
        dataset = dataset.shard(input_context.num_input_pipelines,
                                input_context.input_pipeline_id)

    if is_training:
        # Shuffle the input files
        dataset = dataset.shuffle(buffer_size=_NUM_TRAIN_FILES)

    # Convert to individual records.
    # cycle_length = 10 means that up to 10 files will be read and deserialized in
    # parallel. You may want to increase this number if you have a large number of
    # CPU cores.
    dataset = dataset.interleave(
        tf.data.TFRecordDataset,
        cycle_length=tf.data.experimental.AUTOTUNE,
        num_parallel_calls=tf.data.experimental.AUTOTUNE)

    if is_training and training_dataset_cache:
        # Improve training performance when training data is in remote storage and
        # can fit into worker memory.
        dataset = dataset.cache()

    return process_record_dataset(
        dataset=dataset,
        is_training=is_training,
        batch_size=batch_size,
        im_size=im_size,
        shuffle_buffer=_SHUFFLE_BUFFER,
        parse_record_fn=parse_record_fn,
        num_epochs=num_epochs,
        dtype=dtype,
        datasets_num_private_threads=datasets_num_private_threads,
        drop_remainder=drop_remainder,
        tf_data_experimental_slack=tf_data_experimental_slack,
    )


def _decode_crop_and_flip(image_buffer, bbox, num_channels):
    """Crops the given image to a random part of the image, and randomly flips.

  We use the fused decode_and_crop op, which performs better than the two ops
  used separately in series, but note that this requires that the image be
  passed in as an un-decoded string Tensor.

  Args:
    image_buffer: scalar string Tensor representing the raw JPEG image buffer.
    bbox: 3-D float Tensor of bounding boxes arranged [1, num_boxes, coords]
      where each coordinate is [0, 1) and the coordinates are arranged as
      [ymin, xmin, ymax, xmax].
    num_channels: Integer depth of the image buffer for decoding.

  Returns:
    3-D tensor with cropped image.

  """
    # A large fraction of image datasets contain a human-annotated bounding box
    # delineating the region of the image containing the object of interest.  We
    # choose to create a new bounding box for the object which is a randomly
    # distorted version of the human-annotated bounding box that obeys an
    # allowed range of aspect ratios, sizes and overlap with the human-annotated
    # bounding box. If no box is supplied, then we assume the bounding box is
    # the entire image.
    sample_distorted_bounding_box = tf.image.sample_distorted_bounding_box(
        tf.image.extract_jpeg_shape(image_buffer),
        bounding_boxes=bbox,
        min_object_covered=0.1,
        aspect_ratio_range=[0.75, 1.33],
        area_range=[0.05, 1.0],
        max_attempts=100,
        use_image_if_no_bounding_boxes=True)
    bbox_begin, bbox_size, _ = sample_distorted_bounding_box

    # Reassemble the bounding box in the format the crop op requires.
    offset_y, offset_x, _ = tf.unstack(bbox_begin)
    target_height, target_width, _ = tf.unstack(bbox_size)
    crop_window = tf.stack([offset_y, offset_x, target_height, target_width])

    # Use the fused decode and crop op here, which is faster than each in series.
    cropped = tf.image.decode_and_crop_jpeg(image_buffer,
                                            crop_window,
                                            channels=num_channels)

    # Flip to add a little more random distortion in.
    cropped = tf.image.random_flip_left_right(cropped)
    return cropped


def _central_crop(image, crop_height, crop_width):
    """Performs central crops of the given image list.

  Args:
    image: a 3-D image tensor
    crop_height: the height of the image following the crop.
    crop_width: the width of the image following the crop.

  Returns:
    3-D tensor with cropped image.
  """
    shape = tf.shape(input=image)
    height, width = shape[0], shape[1]

    amount_to_be_cropped_h = (height - crop_height)
    crop_top = amount_to_be_cropped_h // 2
    amount_to_be_cropped_w = (width - crop_width)
    crop_left = amount_to_be_cropped_w // 2
    return tf.slice(image, [crop_top, crop_left, 0],
                    [crop_height, crop_width, -1])


def _mean_image_subtraction(image, means, num_channels):
    """Subtracts the given means from each image channel.

  For example:
    means = [123.68, 116.779, 103.939]
    image = _mean_image_subtraction(image, means)

  Note that the rank of `image` must be known.

  Args:
    image: a tensor of size [height, width, C].
    means: a C-vector of values to subtract from each channel.
    num_channels: number of color channels in the image that will be distorted.

  Returns:
    the centered image.

  Raises:
    ValueError: If the rank of `image` is unknown, if `image` has a rank other
      than three or if the number of channels in `image` doesn't match the
      number of values in `means`.
  """
    if image.get_shape().ndims != 3:
        raise ValueError('Input must be of size [height, width, C>0]')

    if len(means) != num_channels:
        raise ValueError('len(means) must match the number of channels')

    # We have a 1-D tensor of means; convert to 3-D.
    # Note(b/130245863): we explicitly call `broadcast` instead of simply
    # expanding dimensions for better performance.
    means = tf.broadcast_to(means, tf.shape(image))

    return image - means


def _smallest_size_at_least(height, width, resize_min):
    """Computes new shape with the smallest side equal to `smallest_side`.

  Computes new shape with the smallest side equal to `smallest_side` while
  preserving the original aspect ratio.

  Args:
    height: an int32 scalar tensor indicating the current height.
    width: an int32 scalar tensor indicating the current width.
    resize_min: A python integer or scalar `Tensor` indicating the size of
      the smallest side after resize.

  Returns:
    new_height: an int32 scalar tensor indicating the new height.
    new_width: an int32 scalar tensor indicating the new width.
  """
    resize_min = tf.cast(resize_min, tf.float32)

    # Convert to floats to make subsequent calculations go smoothly.
    height, width = tf.cast(height, tf.float32), tf.cast(width, tf.float32)

    smaller_dim = tf.minimum(height, width)
    scale_ratio = resize_min / smaller_dim

    # Convert back to ints to make heights and widths that TF ops will accept.
    new_height = tf.cast(height * scale_ratio, tf.int32)
    new_width = tf.cast(width * scale_ratio, tf.int32)

    return new_height, new_width


def _aspect_preserving_resize(image, resize_min):
    """Resize images preserving the original aspect ratio.

  Args:
    image: A 3-D image `Tensor`.
    resize_min: A python integer or scalar `Tensor` indicating the size of
      the smallest side after resize.

  Returns:
    resized_image: A 3-D tensor containing the resized image.
  """
    shape = tf.shape(input=image)
    height, width = shape[0], shape[1]

    new_height, new_width = _smallest_size_at_least(height, width, resize_min)

    return _resize_image(image, new_height, new_width)


def _resize_image(image, height, width):
    """Simple wrapper around tf.resize_images.

  This is primarily to make sure we use the same `ResizeMethod` and other
  details each time.

  Args:
    image: A 3-D image `Tensor`.
    height: The target height for the resized image.
    width: The target width for the resized image.

  Returns:
    resized_image: A 3-D tensor containing the resized image. The first two
      dimensions have the shape [height, width].
  """
    return tf.compat.v1.image.resize(image, [height, width],
                                     method=tf.image.ResizeMethod.BILINEAR,
                                     align_corners=False)


# added this function from original inception preprocessing -- KDC
def _random_distort_color(image):
    """randomly distort the color of a Tensor image.

      Each color distortion is non-commutative and thus ordering of the color ops
      matters. Ideally we would randomly permute the ordering of the color ops.
      Rather then adding that level of complication, we select a distinct ordering
      of color ops for each preprocessing thread.

      Args:
        image: 3-D Tensor containing single image in [0, 255].
        color_ordering: Python int, a type of distortion (valid values: 0-3).
      Returns:
        3-D Tensor color-distorted image on range [0, 1]
      Raises:
        ValueError: if color_ordering not in [0, 3]
    """
    num_distort_cases = 1
    # scale image from 0 to 255 to 0 to 1 to use distort color
    image = tf.math.scalar_mul(1.0 / 255.0, image)
    distorted_image = _apply_with_random_selector(
        image,
        lambda x, ordering: _distort_color(x, ordering),
        num_cases=num_distort_cases)
    return distorted_image


# added this function from original inception preprocessing -- KDC
def _apply_with_random_selector(x, func, num_cases):
    """Computes func(x, sel), with sel sampled from [0...num_cases-1].

      Args:
        x: input Tensor.
        func: Python function to apply.
        num_cases: Python int32, number of cases to sample sel from.

      Returns:
        The result of func(x, sel), where func receives the value of the
        selector as a python integer, but sel is sampled dynamically.
    """
    sel = tf.random.uniform([], maxval=num_cases, dtype=tf.int32)
    # Pass the real x only to one of the func calls.
    return control_flow_ops.merge([
        func(control_flow_ops.switch(x, tf.equal(sel, case))[1], case)
        for case in range(num_cases)
    ])[0]


# added this function from original inception preprocessing -- KDC
def _distort_color(image, color_ordering=0):
    """Distort the color of a Tensor image.

      Each color distortion is non-commutative and thus ordering of the color ops
      matters. Ideally we would randomly permute the ordering of the color ops.
      Rather then adding that level of complication, we select a distinct ordering
      of color ops for each preprocessing thread.

      Args:
        image: 3-D Tensor containing single image in [0, 1].
        color_ordering: Python int, a type of distortion (valid values: 0-3).
      Returns:
        3-D Tensor color-distorted image on range [0, 1]
      Raises:
        ValueError: if color_ordering not in [0, 3]
    """
    if color_ordering == 0:
        image = tf.image.random_brightness(image, max_delta=32. / 255.)
        image = tf.image.random_saturation(image, lower=0.5, upper=1.5)
    else:
        image = tf.image.random_saturation(image, lower=0.5, upper=1.5)
        image = tf.image.random_brightness(image, max_delta=32. / 255.)

    # The random_* ops do not necessarily clamp.
    return tf.clip_by_value(image, 0.0, 1.0)


def resize_and_crop(image_buffer, output_height, output_width, num_channels):
    """Resize and crop the given image.

  Args:
    image_buffer: scalar string Tensor representing the raw JPEG image buffer.
    output_height: The height of the image after preprocessing.
    output_width: The width of the image after preprocessing.
    num_channels: Integer depth of the image buffer for decoding.

  Returns:
    A resized and cropped image as a numpy array in uint8.
    """
    image = preprocess_image(image_buffer=image_buffer,
                             bbox=None,
                             output_height=output_height,
                             output_width=output_width,
                             num_channels=num_channels,
                             is_training=False,
                             alpha=1.,
                             beta=0.)

    image = tf.cast(image, tf.uint8)
    return np.expand_dims(image.numpy(), axis=0)


def preprocess_image(image_buffer,
                     bbox,
                     output_height,
                     output_width,
                     num_channels,
                     is_training=False,
                     alpha=128.,
                     beta=128.):
    """Preprocesses the given image.

  Preprocessing includes decoding, cropping, and resizing for both training
  and eval images. Training preprocessing, however, introduces some random
  distortion of the image to improve accuracy.

  Args:
    image_buffer: scalar string Tensor representing the raw JPEG image buffer.
    bbox: 3-D float Tensor of bounding boxes arranged [1, num_boxes, coords]
      where each coordinate is [0, 1) and the coordinates are arranged as
      [ymin, xmin, ymax, xmax].
    output_height: The height of the image after preprocessing.
    output_width: The width of the image after preprocessing.
    num_channels: Integer depth of the image buffer for decoding.
    is_training: `True` if we're preprocessing the image for training and
      `False` otherwise.

  Returns:
    A preprocessed image.
  """
    if is_training:
        # For training, we want to randomize some of the distortions.
        image = _decode_crop_and_flip(image_buffer, bbox, num_channels)
        image = _resize_image(image, output_height, output_width)
        # this could improve things but we skip it for now -- KDC
        #image = _random_distort_color(image)
    else:
        # For validation, we want to decode, resize, then just crop the middle.
        image = tf.image.decode_jpeg(image_buffer, channels=num_channels)
        image = _aspect_preserving_resize(image, _RESIZE_MIN)
        image = _central_crop(image, output_height, output_width)

    image.set_shape([output_height, output_width, num_channels])

    # Convert image to float in case the resize didn't
    image = tf.cast(image, tf.float32)

    # Here, we change the mean value to subtract from the image because Akida
    # supports a single beta value for all channels.
    #return _mean_image_subtraction(image, CHANNEL_MEANS, num_channels)
    image = _mean_image_subtraction(image, [beta, beta, beta], num_channels)

    # Divide by alpha
    return tf.math.scalar_mul(1 / alpha, image)


def index_to_label(index):
    """ Function to get an ImageNet label from an index.

  Args:
    index: between 0 and 999


  Returns:
    a string of coma separated labels
  """
    return imagenet_labels[index]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import os

from tensorflow.keras.layers import Input
from tensorflow.keras import Model
from tensorflow.keras.utils import get_file

from ..quantization_blocks import conv_block, separable_conv_block

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/mobilenet/'


def mobilenet_cifar10(input_shape,
                      classes=10,
                      weights=None,
                      weight_quantization=0,
                      activ_quantization=0,
                      input_weight_quantization=None):
    """Instantiates a MobileNet-like model for the "Cifar-10" example.
    This model is based on the MobileNet architecture, mainly with fewer layers.
    The weights and activations are quantized such that it can be converted into
    an Akida model.
    This architecture is inspired by MobileNet.

    Args:
        input_shape (tuple): input shape tuple of the model
        classes (int): number of classes to classify images into
        weights (str): one of `None` (random initialization), 'cifar10' for
            pretrained weights or the path to the weights file to be loaded.
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a.
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization: sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * 'None' implements the same bitwidth as the other weights.
            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.

    Returns:
        tf.keras.model: a quantized Keras model for mobileNet/cifar10
    """

    # Overrides input weight quantization if None
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Check weights
    if not (weights in {None, 'cifar10'} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization) '
                         'or `cifar10` to load weights over http '
                         'or the path to the weights file to be loaded.')

    # Check if Cifar-10 pretrained weights are compatible with given arguments
    if weights == 'cifar10':
        quant_params = (weight_quantization, activ_quantization,
                        input_weight_quantization)
        if classes != 10:
            raise ValueError("If 'weights' is 'cifar10', classes should be 10")
        if quant_params != (4, 4, 8):
            raise ValueError(
                "If 'weights' is 'cifar10', quantization parameters "
                "(weight_quantization, activ_quantization, "
                " input_weight_quantization) must be (4, 4, 8)")
        if input_shape[-1] != 3:
            raise ValueError("If 'weights' is 'cifar10', channel number"
                             " must be 3")
        if input_shape[0] != 32 or input_shape[1] != 32:
            print(
                "If 'weights' is 'cifar10', weights trained for input (32, 32) "
                "will be used")

    img_input = Input(shape=input_shape)

    x = conv_block(img_input,
                   filters=128,
                   name='conv_0',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   weight_quantization=input_weight_quantization,
                   activ_quantization=activ_quantization,
                   add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=128,
                             kernel_size=(3, 3),
                             name='separable_1',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=256,
                             kernel_size=(3, 3),
                             name='separable_2',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=256,
                             kernel_size=(3, 3),
                             name='separable_3',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             pooling='max',
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_4',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_5',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             pooling='max',
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_6',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=512,
                             kernel_size=(3, 3),
                             name='separable_7',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             pooling='max',
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=1024,
                             kernel_size=(3, 3),
                             name='separable_8',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=1024,
                             kernel_size=(3, 3),
                             name='separable_9',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=activ_quantization,
                             add_batchnorm=True)

    x = separable_conv_block(x,
                             filters=classes,
                             kernel_size=(3, 3),
                             name='separable_10',
                             padding='same',
                             use_bias=False,
                             weight_quantization=weight_quantization,
                             activ_quantization=None,
                             pooling='global_avg',
                             add_batchnorm=False)

    model = Model(img_input, x, name='mobilenet_cifar10')

    # Load weights
    if weights == 'cifar10':
        model_name = (f'mobilenet_cifar10_wq{weight_quantization}'
                      f'_aq{activ_quantization}.hdf5')
        weights_path = get_file(fname=model_name,
                                origin=BASE_WEIGHT_PATH + model_name,
                                cache_subdir='models')
        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    return model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

# Keras Imports
from tensorflow.keras.layers import BatchNormalization, ReLU
from tensorflow.keras.layers import MaxPool2D, AvgPool2D, GlobalAvgPool2D

# CNN2SNN Imports
import cnn2snn.quantization_ops as qops
from cnn2snn.quantization_layers import conv2d, separable_conv2d, dense
from cnn2snn.quantization_layers import activationDiscreteRelu


def _get_weight_quantizer(bitwidth):
    """Returns a weight quantizer according to the bitwidth.

    Args:
        bitwidth: (integer) The quantization bitwidth for weights. If
            'bitwidth' is zero, the weights are not quantized (float).

    Returns:
        a WeightQuantizer object.
    """
    if bitwidth == 0:
        return qops.WeightFloat()
    elif isinstance(bitwidth, int) and bitwidth > 0:
        if bitwidth <= 2:
            # For very low bitwidth, we must use a lower theshold otherwise we
            # set too many weights to zero
            return qops.WeightQuantizer(threshold=1, bitwidth=bitwidth)
        else:
            return qops.WeightQuantizer(threshold=3, bitwidth=bitwidth)
    else:
        raise ValueError("'bitwidth' argument must be a positive integer, "
                         "or zero for float weights.")


def _add_pooling_layer(x, pooling_type, pool_size, padding, layer_base_name):
    """Add a pooling layer in the graph.

    From an input tensor 'x', the function returns the output tensor after
    a pooling layer defined by 'pooling_type'.

    Args:
        x (tf.Tensor): the input tensor
        pooling_type (str): type of pooling among the following: 'max',
            'avg' or 'global_avg'.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        padding (str): one of "valid" or "same" (case-insensitive).
        layer_base_name (str): base name for the pooling layer.

    Returns:
        tf.Tensor: an output tensor after pooling
    """
    if pooling_type == 'max':
        return MaxPool2D(pool_size=pool_size,
                         padding=padding,
                         name=layer_base_name + '_maxpool')(x)
    elif pooling_type == 'avg':
        return AvgPool2D(pool_size=pool_size,
                         padding=padding,
                         name=layer_base_name + '_avgpool')(x)
    elif pooling_type == 'global_avg':
        return GlobalAvgPool2D(name=layer_base_name + '_global_avg')(x)
    else:
        raise ValueError("'pooling_type' argument must be max', "
                         "'avg' or 'global_avg'.")


def _add_activation_layer(x, bitwidth, layer_name):
    """Add an activation layer in the graph.

    From an input tensor 'x', the function returns the output tensor after
    an activation layer defined by 'bitwidth'.

    Args:
        x (tf.Tensor): the input tensor
        bitwidth (int): the quantization bitwidth for the ReLU6
            activation. If 'bitwidth' is zero, the activations are not
            quantized (float ReLU6).
        layer_name (str): This name will be used as the name of the added
            activation layer.

    Returns:
        tf.Tensor: an output tensor after activation.
    """
    if bitwidth == 0:
        return ReLU(6., name=layer_name)(x)
    elif isinstance(bitwidth, int) and bitwidth > 0:
        return activationDiscreteRelu(bitwidth=bitwidth, name=layer_name)(x)
    else:
        raise ValueError("'bitwidth' argument must be a positive integer, "
                         "or zero for float activation.")


def conv_block(inputs,
               filters,
               kernel_size,
               weight_quantization=0,
               activ_quantization=0,
               pooling=None,
               pool_size=(2, 2),
               add_batchnorm=False,
               **kwargs):
    """Adds a quantized convolutional layer with optional layers in the
    following order: max pooling, batch normalization, quantized activation.

    Args:
        inputs (tf.Tensor): input tensor of shape `(rows, cols, channels)`
        filters (int): the dimensionality of the output space
            (i.e. the number of output filters in the convolution).
        kernel_size (int or tuple of 2 integers): specifying the
            height and width of the 2D convolution kernel.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        weight_quantization (int): quantization bitwidth for weights
            (usually 2, 4 or 8). For float weights (no quantization), set the
            value to zero.
        activ_quantization (int) quantization bitwidth for activation
            (usually 1, 2 or 4). For a float activation (ReLU 6), set the
            value to zero. For no activation, set it to None.
        pooling (str): add a pooling layer of type 'pooling' among the
            values 'max', 'avg', 'global_max' or 'global_avg', with pooling
            size set to pool_size. If 'None', no pooling will be added.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        add_batchnorm (bool): add a tf.keras.BatchNormalization layer
        **kwargs: arguments passed to the tf.keras.Conv2D layer, such as
            strides, padding, use_bias, weight_regularizer, etc.

    Returns:
        tf.Tensor: output tensor of conv2D block.
    """
    if 'activation' in kwargs.keys() and kwargs['activation']:
        raise ValueError(
            "Keyword argument 'activation' in conv_block must be None.")
    if 'strides' in kwargs.keys():
        if kwargs['strides'] != 1 and kwargs['strides'] != (1, 1):
            print("WARNING: Keyword argument 'strides' is not supported in "
                  "conv_block except for the first layer.")
    if 'dilation_rate' in kwargs.keys():
        if kwargs['dilation_rate'] != 1 and kwargs['dilation_rate'] != (1, 1):
            raise ValueError("Keyword argument 'dilation_rate' is not "
                             "supported in conv_block.")

    weight_quantizer = _get_weight_quantizer(weight_quantization)
    conv_layer = conv2d(filters,
                        kernel_size,
                        modifier=weight_quantizer,
                        **kwargs)

    x = inputs
    x = conv_layer(x)

    if pooling:
        x = _add_pooling_layer(x, pooling, pool_size, conv_layer.padding,
                               conv_layer.name)

    if add_batchnorm:
        x = BatchNormalization(name=conv_layer.name + '_BN')(x)

    if activ_quantization is not None:
        x = _add_activation_layer(x, activ_quantization,
                                  conv_layer.name + '_relu')

    return x


def separable_conv_block(inputs,
                         filters,
                         kernel_size,
                         weight_quantization=0,
                         activ_quantization=0,
                         pooling=None,
                         pool_size=(2, 2),
                         add_batchnorm=False,
                         **kwargs):
    """Adds a quantized separable convolutional layer with optional layers in
    the following order: global average pooling, max pooling, batch
    normalization, quantized activation.

    Args:
        inputs (tf.Tensor): input tensor of shape `(height, width, channels)`
        filters (int): the dimensionality of the output space
            (i.e. the number of output filters in the pointwise convolution).
        kernel_size (int or tuple of 2 integers): specifying the
            height and width of the 2D convolution window. Can be a single
            integer to specify the same value for all spatial dimensions.
        weight_quantization (int): quantization bitwidth for weights
            (usually 2, 4 or 8). For float weights (no quantization), set the
            value to zero.
        activ_quantization (int): quantization bitwidth for activation
            (usually 1, 2 or 4). For a float activation (ReLU 6), set the
            value to zero. For no activation, set it to None.
        pooling (str): add a pooling layer of type 'pooling' among the
            values 'max', 'avg', 'global_max' or 'global_avg', with pooling
            size set to pool_size. If 'None', no pooling will be added.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        add_batchnorm (bool): add a tf.keras.BatchNormalization layer
        **kwargs: arguments passed to the tf.keras.SeparableConv2D layer,
            such as strides, padding, use_bias, etc.

    Returns:
        tf.Tensor: output tensor of separable conv block.
    """
    if 'activation' in kwargs.keys() and kwargs['activation']:
        raise ValueError("Keyword argument 'activation' in separable_conv_block"
                         " must be None.")
    if 'strides' in kwargs.keys():
        if kwargs['strides'] != 1 and kwargs['strides'] != (1, 1):
            raise ValueError("Keyword argument 'strides' is not supported in "
                             "separable_conv_block.")
    if 'dilation_rate' in kwargs.keys():
        if kwargs['dilation_rate'] != 1 and kwargs['dilation_rate'] != (1, 1):
            raise ValueError("Keyword argument 'dilation_rate' is not "
                             "supported in separable_conv_block.")
    if 'depth_multiplier' in kwargs.keys():
        if kwargs['depth_multiplier'] != 1:
            raise ValueError("Keyword argument 'depth_multiplier' is not "
                             "supported in separable_conv_block.")

    weight_quantizer = _get_weight_quantizer(weight_quantization)
    separable_layer = separable_conv2d(filters,
                                       kernel_size,
                                       modifier=weight_quantizer,
                                       **kwargs)

    x = inputs
    x = separable_layer(x)

    if pooling:
        x = _add_pooling_layer(x, pooling, pool_size, separable_layer.padding,
                               separable_layer.name)

    if add_batchnorm:
        x = BatchNormalization(name=separable_layer.name + '_BN')(x)

    if activ_quantization is not None:
        x = _add_activation_layer(x, activ_quantization,
                                  separable_layer.name + '_relu')

    return x


def dense_block(inputs,
                units,
                weight_quantization=0,
                activ_quantization=0,
                add_batchnorm=False,
                **kwargs):
    """Adds a quantized dense layer with optional layers in the following
    order: batch normalization, quantized activation.

    Args:
        inputs (tf.Tensor): Input tensor of shape `(rows, cols, channels)`
        units (int): dimensionality of the output space
        weight_quantization (int): quantization bitwidth for weights
            (usually 2, 4 or 8). For float weights (no quantization), set the
            value to zero.
        activ_quantization (int): quantization bitwidth for activation
            (usually 1, 2 or 4). For a float activation (ReLU 6), set the
            value to zero. For no activation, set it to None.
        add_batchnorm (bool): add a tf.keras.BatchNormalization layer
        **kwargs: arguments passed to the tf.keras.Dense layer, such as
            use_bias, kernel_initializer, weight_regularizer, etc.

    Returns:
        tf.Tensor: output tensor of the dense block.
    """
    if 'activation' in kwargs.keys() and kwargs['activation']:
        raise ValueError("Keyword argument 'activation' in dense_block"
                         " must be None.")

    weight_quantizer = _get_weight_quantizer(weight_quantization)
    dense_layer = dense(units, modifier=weight_quantizer, **kwargs)

    x = inputs
    x = dense_layer(x)

    if add_batchnorm:
        x = BatchNormalization(name=dense_layer.name + '_BN')(x)

    if activ_quantization is not None:
        x = _add_activation_layer(x, activ_quantization,
                                  dense_layer.name + '_relu')

    return x



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

# System imports
import os
import warnings

# Tensorflow/Keras Imports
from tensorflow.keras.utils import get_file
from tensorflow.keras import backend, Model
from tensorflow.keras.layers import (Input, BatchNormalization, MaxPool2D,
                                     GlobalAvgPool2D, Reshape, Dropout,
                                     Activation)

# cnn2snn Imports
from cnn2snn.quantization_ops import WeightQuantizer, WeightFloat
from cnn2snn.quantization_layers import conv2d, separable_conv2d, dense

# Local utils
from ..quantization_blocks import _add_activation_layer, _get_weight_quantizer

# Locally fixed config options
# The number of neurons in the penultimate dense layer
# This layer has binary output spikes, and could be a bottleneck
# if care isn't taken to ensure enough info capacity
NUM_SPIKING_NEURONS = 256

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/convtiny/'


def _depthwise_block(inputs, filters, name, add_maxpool, weight_quantizer,
                     activ_quantization):
    """Returns a depthwise block.

    Args:
        inputs (tuple): 4D tensor previous layer output shape
        filters (int): number of neurons
        name (str): the depthwise block name
        add_maxpool (bool): add MaxPool2D block
        weight_quantizer (int):
        activ_quantization (int): sets all activations in the model

    Returns:
        tuple: the output shape as 4D tensor
    """
    x = separable_conv2d(filters=filters,
                         kernel_size=(3, 3),
                         use_bias=False,
                         padding='same',
                         name=name,
                         modifier=weight_quantizer)(inputs)

    if add_maxpool:
        x = MaxPool2D(pool_size=(2, 2), padding='same',
                      name=name + '_maxpool')(x)
    x = BatchNormalization(name=name + '_BN')(x)
    x = _add_activation_layer(x, activ_quantization, name + '_relu')

    return x


def convtiny_dvs(input_shape,
                 weights=None,
                 classes=None,
                 weight_quantization=0,
                 activ_quantization=0):
    """Instantiates a CNN for the "IBM DVS Gesture" example.

    Args:
        input_shape (tuple): input shape tuple of the model
        weights (str): one of `None` (random initialization), 'dvs_gesture' for
            pretrained weights or the path to the weights file to be loaded.
        classes (int): number of classes to classify images into.
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '1' through '8' implements n-bit weights where n is from 1-8 bits.

    Returns:
        tf.keras.model: a quantized Keras convolutional model for DVS Gesture.
    """
    # Check weights
    if not (weights in {None, 'dvs_gesture'} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization) '
                         'or `dvs_gesture` to load weights over http '
                         'or the path to the weights file to be loaded.')

    # Check if dvs_gesture pretrained weights are compatible with given arguments
    if weights == 'dvs_gesture':
        quant_params = (weight_quantization, activ_quantization)
        if quant_params != (2, 4):
            raise ValueError("If 'weights' is 'dvs_gesture', quantization"
                             " parameters (weight_quantization, "
                             " activ_quantization, input_weight_quantization)"
                             " must be (2, 4); "
                             f" got quantization parameters={quant_params}")
        if input_shape != (64, 64, 2):
            raise ValueError(
                "If 'weights' is 'dvs_gesture', input shape must be "
                f" (64,64,2) ; got input_shape={input_shape}")

    img_input = Input(input_shape)

    # Set quantization
    weight_quantizer = _get_weight_quantizer(weight_quantization)

    num_blocks = 4
    x = img_input
    for block in range(num_blocks):
        x = conv2d(filters=16 * 2**block,
                   kernel_size=(3, 3),
                   use_bias=False,
                   strides=(1, 1),
                   padding='same',
                   name='conv_' + str(block),
                   modifier=weight_quantizer)(x)
        if block < (num_blocks - 2):
            x = MaxPool2D(pool_size=(3, 3),
                          padding='same',
                          name='block_' + str(block) + '_maxpool')(x)
        elif block == (num_blocks - 1):
            x = GlobalAvgPool2D(name='global_avg')(x)
        elif block >= (num_blocks - 2):
            x = MaxPool2D(pool_size=(2, 2),
                          padding='same',
                          name='block_' + str(block) + '_maxpool')(x)

        x = BatchNormalization(name='block_' + str(block) + '_BN')(x)
        x = _add_activation_layer(x, activ_quantization,
                                  'block_' + str(block) + '_relu')

    bm_outshape = (1, 1, 128)

    x = Reshape(bm_outshape, name='reshape_1')(x)
    x = Dropout(1e-3, name='dropout')(x)

    aq = 0
    if weight_quantization > 0:
        # Assume that if we're quantizing weights, we want the Spiking Layer to
        # generate spikes
        aq = 1
        # Assume that if we're quantizing weights, we want the Dense layer to
        # be HW compatible, => WQ=2
        fc_weight_quantizer = _get_weight_quantizer(2)
    else:
        aq = activ_quantization
        fc_weight_quantizer = _get_weight_quantizer(weight_quantization)

    x = _depthwise_block(x, NUM_SPIKING_NEURONS, 'spiking_layer', False,
                         weight_quantizer, aq)

    x = dense(classes, modifier=fc_weight_quantizer, use_bias=False)(x)
    x = Activation('softmax', name='act_softmax')(x)
    x = Reshape((classes,), name='reshape_2')(x)

    model = Model(inputs=img_input, outputs=x, name='dvs_network')

    # Load weights.
    if weights == 'dvs_gesture':
        model_name = (f'convtiny_dvs_gesture_wq{weight_quantization}'
                      f'_aq{activ_quantization}.h5')
        weights_path = get_file(fname=model_name,
                                origin=BASE_WEIGHT_PATH + model_name,
                                cache_subdir='models')
        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    return model



#!/usr/bin/env python
"""MobileNet v1 models for Keras and Akida
MobileNet is a general architecture and can be used for multiple use cases.
(general description from keras.applications still included immediately below).

AKIDA
This specific version includes parameter options to generate a mobilenet version
with
- overall architecture compatible with Akida (conv stride 2 replaced with max pool)
- options to binarize or ternarize both weights and activations
- options to customize (including remove) batch norm + activation layers
between the depthwise and pointwise components of the depthwise_conv_block
- different initialization options

General Mobilenet V1 Notes:
Depending on the use case, it can use different input layer size and
different width factors. This allows different width models to reduce
the number of multiply-adds and thereby
reduce inference cost on mobile devices.
MobileNets support any input size greater than 32 x 32, with larger image sizes
offering better performance.
The number of parameters and number of multiply-adds
can be modified by using the `alpha` parameter,
which increases/decreases the number of filters in each layer.
By altering the image size and `alpha` parameter,
all 16 models from the paper can be built, with ImageNet weights provided.
The paper demonstrates the performance of MobileNets using `alpha` values of
1.0 (also called 100 % MobileNet), 0.75, 0.5 and 0.25.
For each of these `alpha` values, weights for 4 different input image sizes
are provided (224, 192, 160, 128).
The following table describes the size and accuracy of the 100% MobileNet
on size 224 x 224:
----------------------------------------------------------------------------
Width Multiplier (alpha) | ImageNet Acc |  Multiply-Adds (M) |  Params (M)
----------------------------------------------------------------------------
|   1.0 MobileNet-224    |    70.6 %     |        529        |     4.2     |
|   0.75 MobileNet-224   |    68.4 %     |        325        |     2.6     |
|   0.50 MobileNet-224   |    63.7 %     |        149        |     1.3     |
|   0.25 MobileNet-224   |    50.6 %     |        41         |     0.5     |
----------------------------------------------------------------------------
The following table describes the performance of
the 100 % MobileNet on various input sizes:
------------------------------------------------------------------------
      Resolution      | ImageNet Acc | Multiply-Adds (M) | Params (M)
------------------------------------------------------------------------
|  1.0 MobileNet-224  |    70.6 %    |        529        |     4.2     |
|  1.0 MobileNet-192  |    69.1 %    |        529        |     4.2     |
|  1.0 MobileNet-160  |    67.2 %    |        529        |     4.2     |
|  1.0 MobileNet-128  |    64.4 %    |        529        |     4.2     |
------------------------------------------------------------------------
The weights for all 16 models are obtained and translated
from TensorFlow checkpoints found at
https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet_v1.md
# Reference
- [MobileNets: Efficient Convolutional Neural Networks for
   Mobile Vision Applications](https://arxiv.org/pdf/1704.04861.pdf))
"""

import os
import warnings

# Tensorflow/Keras Imports
import tensorflow.keras.utils as keras_utils
from tensorflow.keras import Model
from tensorflow.keras.utils import get_file
from tensorflow.keras.layers import (Input, BatchNormalization, MaxPool2D,
                                     GlobalAvgPool2D, Reshape, Dropout,
                                     Activation)

# cnn2snn Imports
from cnn2snn import (WeightQuantizer, WeightFloat, QuantizedConv2D,
                     QuantizedSeparableConv2D)

# Local utils
from ..quantization_blocks import _add_activation_layer

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/mobilenet/'


def _obtain_input_shape(input_shape, default_size, min_size, include_top):
    """Internal utility to compute/validate a model's input shape.

    Args:
        input_shape: either None (will return the default model input shape),
            or a user-provided shape to be validated.
        default_size: default input width/height for the model.
        min_size: minimum input width/height accepted by the model.
        include_top: whether the model is expected to
            be linked to a classifier via a Flatten layer.

    Returns:
        tuple of integers: input shape (may include None entries).

    Raises:
        ValueError: in case of invalid argument values.
    """
    if input_shape and len(input_shape) == 3:
        if input_shape[-1] not in {1, 3}:
            warnings.warn('This model usually expects 1 or 3 input channels. '
                          'However, it was passed an input_shape with ' +
                          str(input_shape[-1]) + ' input channels.')
        default_shape = (default_size, default_size, input_shape[-1])
    else:
        default_shape = (default_size, default_size, 3)

    if input_shape:
        if input_shape is not None:
            if len(input_shape) != 3:
                raise ValueError(
                    '`input_shape` must be a tuple of three integers.')
            if ((input_shape[0] is not None and input_shape[0] < min_size) or
                (input_shape[1] is not None and input_shape[1] < min_size)):
                raise ValueError('Input size must be at least ' +
                                 str(min_size) + 'x' + str(min_size) +
                                 '; got `input_shape=' + str(input_shape) + '`')
    else:
        if include_top:
            input_shape = default_shape
        else:
            input_shape = (None, None, 3)
    if include_top:
        if None in input_shape:
            raise ValueError('If `include_top` is True, '
                             'you should specify a static `input_shape`. '
                             'Got `input_shape=' + str(input_shape) + '`')
    return input_shape


def _depthwise_block(inputs, filters, name, add_maxpool, weight_quantizer,
                     activ_quantization):

    x = QuantizedSeparableConv2D(filters=filters,
                                 kernel_size=(3, 3),
                                 use_bias=False,
                                 padding='same',
                                 name=name,
                                 quantizer=weight_quantizer)(inputs)

    if add_maxpool:
        x = MaxPool2D(pool_size=(2, 2), padding='same',
                      name=name + '_maxpool')(x)
    x = BatchNormalization(name=name + '_BN')(x)
    x = _add_activation_layer(x, activ_quantization, name + '_relu')

    return x


def _get_weight_quantizer(bitwidth):
    """Returns a weight quantizer according to the bitwidth.

    # Args
        bitwidth (int): The quantization bitwidth for weights. If
            'bitwidth' is zero, the weights are not quantized (float).

    # Returns
        WeightQuantizer: a weight quantizer object.
    """
    if bitwidth == 0:
        return WeightFloat()
    elif isinstance(bitwidth, int) and bitwidth > 0:
        return WeightQuantizer(threshold=1, bitwidth=bitwidth)
    else:
        raise ValueError("'bitwidth' argument must be a positive integer, "
                         "or zero for float weights.")


def mobilenet_imagenet(input_shape=None,
                       alpha=1.0,
                       dropout=1e-3,
                       include_top=True,
                       weights=None,
                       pooling=None,
                       classes=1000,
                       weight_quantization=0,
                       activ_quantization=0,
                       input_weight_quantization=None):
    """Instantiates the MobileNet architecture.

    Args:
        input_shape (tuple): optional shape tuple.
        alpha (float): controls the width of the model.

            * If `alpha` < 1.0, proportionally decreases the number of filters
              in each layer.
            * If `alpha` > 1.0, proportionally increases the number of filters
              in each layer.
            * If `alpha` = 1, default number of filters from the paper are used
              at each layer.
        dropout (float): dropout rate
        include_top (bool): whether to include the fully-connected
            layer at the top of the model.
        weights (str): one of `None` (random initialization), 'imagenet' for
             pretrained weights or the path to the weights file to be loaded.
        pooling (str): Optional pooling mode for feature extraction
            when `include_top` is `False`.

            * `None` means that the output of the model will be the 4D tensor
              output of the last convolutional block.
            * `avg` means that global average pooling will be applied to the
              output of the last convolutional block, and thus the output of the
              model will be a 2D tensor.
        classes (int): optional number of classes to classify images
            into, only to be specified if `include_top` is True.
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization: sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.

    Returns:
        A Keras model instance.

    Raises:
        ValueError: in case of invalid argument for `weights`,
            or invalid input shape.
    """
    # check if overrides have been provided and override
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Check weights
    if not (weights in {None, 'imagenet'} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization) '
                         'or `imagenet` to load weights over http '
                         'or the path to the weights file to be loaded.')

    # Check if imagenet pretrained weights are compatible with given arguments
    if weights == 'imagenet':
        quant_params = (weight_quantization, activ_quantization,
                        input_weight_quantization)
        if classes != 1000:
            raise ValueError(
                "If 'weights' is 'imagenet', classes should be 1000")
        if quant_params != (4, 4, 8) and quant_params != (0, 0, 0):
            raise ValueError(
                "If 'weights' is 'imagenet', quantization parameters "
                "(weight_quantization, activ_quantization, "
                " input_weight_quantization) must be (4, 4, 8) or"
                " (0, 0, 0)")
        if input_shape[-1] != 3:
            raise ValueError("If 'weights' is 'imagenet', channel number"
                             " must be 3")
        if input_shape[0] != 224 or input_shape[1] != 224:
            print(
                "If 'weights' is 'imagenet', weights trained for input (224, 224) "
                "will be used")

    # Determine proper input shape and default size.
    if input_shape is None:
        default_size = 224
    else:
        rows = input_shape[0]
        cols = input_shape[1]

        if rows == cols and rows in [128, 160, 192, 224]:
            default_size = rows
        else:
            default_size = 224

    input_shape = _obtain_input_shape(input_shape,
                                      default_size=default_size,
                                      min_size=32,
                                      include_top=include_top)

    rows = input_shape[0]
    cols = input_shape[1]

    img_input = Input(shape=input_shape)

    weight_quantizer = _get_weight_quantizer(weight_quantization)
    weight_quantizer_fl = _get_weight_quantizer(input_weight_quantization)

    x = QuantizedConv2D(filters=int(32 * alpha),
                        kernel_size=(3, 3),
                        use_bias=False,
                        strides=(2, 2),
                        padding='same',
                        name='conv_0',
                        quantizer=weight_quantizer_fl)(img_input)
    x = BatchNormalization(name='conv_0_BN')(x)
    x = _add_activation_layer(x, activ_quantization, 'conv_0_relu')

    x = _depthwise_block(x, int(64 * alpha), 'separable_1', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(128 * alpha), 'separable_2', True,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(128 * alpha), 'separable_3', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(256 * alpha), 'separable_4', True,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(256 * alpha), 'separable_5', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(512 * alpha), 'separable_6', True,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(512 * alpha), 'separable_7', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(512 * alpha), 'separable_8', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(512 * alpha), 'separable_9', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(512 * alpha), 'separable_10', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(512 * alpha), 'separable_11', False,
                         weight_quantizer, activ_quantization)

    x = _depthwise_block(x, int(1024 * alpha), 'separable_12', True,
                         weight_quantizer, activ_quantization)

    # Last separable layer with global pooling
    x = QuantizedSeparableConv2D(filters=int(1024 * alpha),
                                 kernel_size=(3, 3),
                                 use_bias=False,
                                 padding='same',
                                 name='separable_13',
                                 quantizer=weight_quantizer)(x)

    if include_top or pooling == 'avg':
        x = GlobalAvgPool2D(name='separable_13_global_avg')(x)
    x = BatchNormalization(name='separable_13_BN')(x)
    x = _add_activation_layer(x, activ_quantization, 'separable_13_relu')

    if include_top:
        shape = (1, 1, int(1024 * alpha))

        x = Reshape(shape, name='reshape_1')(x)
        x = Dropout(dropout, name='dropout')(x)

        x = QuantizedSeparableConv2D(filters=classes,
                                     kernel_size=(3, 3),
                                     use_bias=False,
                                     padding='same',
                                     name='separable_14',
                                     quantizer=weight_quantizer)(x)
        x = Activation('softmax', name='act_softmax')(x)
        x = Reshape((classes,), name='reshape_2')(x)

    # Create model.
    model = Model(img_input, x, name='mobilenet_%0.2f_%s' % (alpha, rows))
    # Load weights.
    if weights == 'imagenet':
        if include_top:
            include_top_file_name = ''
        else:
            include_top_file_name = '_no_top'
        if input_weight_quantization == 0:
            model_name = (f'mobilenet_imagenet{include_top_file_name}.h5')
        else:
            model_name = (f'mobilenet_imagenet_wq{weight_quantization}'
                          f'_aq{activ_quantization}{include_top_file_name}.h5')
        weights_path = get_file(fname=model_name,
                                origin=BASE_WEIGHT_PATH + model_name,
                                cache_subdir='models')
        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    return model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import os

from tensorflow.keras import Model
from tensorflow.keras.utils import get_file

from .preprocessing import load_labels_mapping, load_columns_type
from ..tabular_data.tabular_data import create_tse, create_mlp

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/tse_mlp/'


def tse_mlp_cse2018(numerical_columns=None,
                    categorical_columns=None,
                    weights=None,
                    weight_quantization=0,
                    activ_quantization=0,
                    input_weight_quantization=None):
    """Instantiates a model composed of a trainable spike encoder and a
    multilayer perceptron for the tabular data example on CSE-CIC-IDS-2018
    dataset.

    Args:
        numerical_columns (list): list of numerical column names. Will use
            preprocessed data when set to `None`.
        categorical_columns (dict): dictionary of categorical column names
            indexing the list of vocabulary for the column. Will use
            preprocessed data when set to `None`.
        weights (str): one of `None` (random initialization), 'cse2018' for
            pretrained weights or the path to the weights file to be loaded.
        weight_quantization (int): sets MLP weights in the model to have a
            particular quantization bitwidth

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '1' through '8' implements n-bit weights where n is from 1-8 bits.
        input_weight_quantization: sets weight quantization in the TSE last
            layer. Defaults to weight_quantization value.

            * 'None' implements the same bitwidth as the other weights.
            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
    Returns:
        tf.keras.Model: a Keras model for tabular data on CSE-CIC-IDS-2018
    """
    # Overrides input weight quantization if None
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Check weights
    if not (weights in {None, 'cse2018'} or os.path.exists(weights)):
        raise ValueError('The `weights` argument should be either '
                         '`None` (random initialization) '
                         'or `cse2018` to load weights over http '
                         'or the path to the weights file to be loaded.')

    # Check if cse2018 pretrained weights are compatible with given arguments
    if weights == 'cse2018':
        quant_params = (weight_quantization, activ_quantization,
                        input_weight_quantization)
        if quant_params != (2, 4, 0):
            raise ValueError(
                "If 'weights' is 'cse2018', quantization parameters "
                "(weight_quantization, activ_quantization, "
                "input_weight_quantization) must be (2, 4, 0); "
                f" got quantization parameters={quant_params}")
        cols_params = (numerical_columns, categorical_columns)
        if cols_params != (None, None):
            raise ValueError(
                "If 'weights' is 'cse2018', columns parameters "
                "(numerical_columns, categorical_columns) must be (None, None)")

    labels_mappings = load_labels_mapping()
    n_classes = len(labels_mappings)

    if numerical_columns == None or categorical_columns == None:
        preprocessed_numerical_columns, preprocessed_categorical_columns = load_columns_type(
        )

    if numerical_columns == None:
        numerical_columns = preprocessed_numerical_columns
    if categorical_columns == None:
        categorical_columns = preprocessed_categorical_columns

    mlp_units_list = [128, 128]

    inputs, tse_output = create_tse(
        numerical_columns,
        categorical_columns,
        units=256,
        input_weight_quantization=input_weight_quantization,
        activ_quantization=activ_quantization)
    mlp_output = create_mlp(tse_output,
                            mlp_units_list,
                            n_classes,
                            weight_quantization=weight_quantization,
                            activ_quantization=activ_quantization)
    model = Model(inputs=inputs, outputs=mlp_output)

    # Load weights.
    if weights == 'cse2018':
        model_name = (f'tse_mlp_cse2018_wq{weight_quantization}'
                      f'_aq{activ_quantization}'
                      f'_iq{input_weight_quantization}.hdf5')
        weights_path = get_file(fname=model_name,
                                origin=BASE_WEIGHT_PATH + model_name,
                                cache_subdir='models')
        model.load_weights(weights_path)
    elif weights is not None:
        model.load_weights(weights)

    return model



from os import path
from setuptools import setup, find_packages

# Read the contents of the README file
directory = path.abspath(path.dirname(__file__))
with open(path.join(directory, 'README'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='akida_models',
      version='1.0.1',
      description='Akida Models',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Corentin Friedrich',
      author_email='cfriedrich@brainchip.com',
      url='https://doc.brainchipinc.com',
      license='Apache 2.0',
      license_files=['LICENSE', 'LICENSE.3rdparty'],
      packages=find_packages(),
      install_requires=['cnn2snn>=1.8.0'],
      python_requires='>=3.6')



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
 Toolset to load IBM DVS Gesture preprocessed dataset.

Brainchip information:
        - AER2.0 specifications:
            https://inivation.com/support/software/fileformat/#bit-30-12
        - AER3.1 specifications:
            https://inivation.com/support/software/fileformat/#aedat-31
"""

import struct
import os
import re
import numpy as np

import sys
import tarfile
import urllib


def _parse_version_string(version_string):
    """Analyze string version from AEDAT file.

    Args:
        version_string (str): value contains file version.

    Returns:
        int: document major version.
    """
    v3 = re.compile('.*#!AER-DAT3.*')
    if v3.match(str(version_string)):
        return 3
    else:
        raise ValueError("Unrecognized or unsupported version: %s" %
                         str(version_string).strip(),
                         file=sys.stderr)


def _skip_ascii_header(file, version):
    """Going through header without reading data.

    Args:
        file (obj): opened AEDAT file
        version (int): AEDAT major version

    Returns:
        int: End of header bytes position.
    """
    bytes_pos = 0  # bytes position
    if version == 3:
        line = file.readline()
        bytes_pos += len(line)
        while line.startswith(b'#'):
            if line == b'#!END-HEADER\r\n':
                break
            else:
                line = file.readline()
                bytes_pos += len(line)
    else:
        raise ValueError("Unsupported version: %d" % version)
    return bytes_pos


def _read_event_v3(file, bytes_pos, trial_times, subj_num, light_cond):
    """Read event in AEDAT v3 format.

    Args:
        file (obj): opened AEDAT file.
        bytes_pos (int): starting events header position.
        trial_times (array): start/stop times for gestures from CSV label file.
        subj_num (int): subject id.
        light_cond (str): lighting condition.
    Returns:
        list: readed data are stored in dict and added to the returned list.
    """
    file.seek(bytes_pos)  # seek to end of header

    # init trials values and reference timestamps
    trials = []
    trial_ind = 1
    num_trials = len(trial_times)
    start_time = int(trial_times[trial_ind][1])
    stop_time = int(trial_times[trial_ind][2])

    ev_list = []
    ev_times = []
    while True:
        # events header part
        header = file.read(28)
        if not header or len(header) == 0:
            break

        # read header
        type = struct.unpack('H', header[0:2])[0]
        source = struct.unpack('H', header[2:4])[0]
        size = struct.unpack('I', header[4:8])[0]
        offset = struct.unpack('I', header[8:12])[0]
        tsoverflow = struct.unpack('I', header[12:16])[0]
        capacity = struct.unpack('I', header[16:20])[0]
        number = struct.unpack('I', header[20:24])[0]
        valid = struct.unpack('I', header[24:28])[0]

        data_length = capacity * size
        data = file.read(data_length)
        counter = 0

        if type == 1:
            while data[counter:counter + size]:
                # unpack data from bytes
                aer_data = struct.unpack('I', data[counter:counter + 4])[0]
                timestamp = struct.unpack(
                    'I', data[counter + 4:counter + 8])[0] | tsoverflow << 31
                x = (aer_data >> 17) & 0x00007FFF
                y = (aer_data >> 2) & 0x00007FFF
                pol = (aer_data >> 1) & 0x00000001
                counter = counter + size

                if timestamp >= start_time:
                    if timestamp < stop_time:
                        # Collect the events as a list (coordinates and times
                        # separately)
                        ev_list.append((x, y, pol))
                        ev_times.append(timestamp)
                    else:
                        # Once stop_time is reached,
                        # Collect the event lists and accompanying details as a
                        # dictionary
                        trial = {
                            'label': trial_times[trial_ind][0],
                            'subject': subj_num,
                            'lighting': light_cond,
                            'events': np.array(ev_list, dtype=np.uint8),
                            'ev_times': np.array(ev_times) - ev_times[0]
                        }
                        trials.append(trial)
                        # Then reset other values ready to collect the next
                        # trial
                        ev_list = []
                        ev_times = []
                        trial_ind += 1
                        if trial_ind < num_trials:
                            start_time = int(trial_times[trial_ind][1])
                            if start_time < stop_time:
                                print("Error on subject %d, with %s condition" %
                                      (subj_num, light_cond))
                                raise Exception(
                                    "New start time earlier than previous block stop time"
                                )
                            stop_time = int(trial_times[trial_ind][2])
                        else:
                            return trials
        else:
            # non-polarity event packet, not implemented
            raise NotImplementedError("Non polarity event packet not supported")
            pass
    return


def get_gesture_data(datafile, labelfile):
    """Creates gesture data dictionary from v3.1 aedat file.

    Args:
        datafile (str): absolute path to aedat file
        labelfile (int): absolute path to related label file.

    Returns:
        list: loaded data stored in multiple dict with gesture id, subject id,
            lighting condition, events and events time.
    """
    with open(labelfile) as f:
        # Get class labels and start + stop times from the _labels.csv file
        trial_times = [line.split(',') for line in f]

    if os.path.isfile(datafile):
        # Get subject id and lightning condition from filename
        subj_num = int(os.path.basename(datafile)[4:6])
        lighting_condition = os.path.basename(datafile)[7:-6]

        aef = open(datafile, 'rb')
        version = _parse_version_string(aef.readline())
        aef.seek(0, 0)  # rewind
        bytes_pos = _skip_ascii_header(aef, version)
        if version == 3:
            data = _read_event_v3(aef, bytes_pos, trial_times, subj_num,
                                  lighting_condition)
        else:
            raise ValueError("Unsupported version: %d" % version)
    return data


def generate_dataset(datadir, subject_list):
    """Generates data set for a subject id list.

    Args:
        datadir (str): filepath where aedat files are stored.
        subject_list (list): subject id list to compose the train/test set.

    Returns:
        list: dataset filled with all events data.
    """
    dataset = []
    for subj in subject_list:
        import glob
        datafile = glob.glob(
            os.path.join(datadir, 'DvsGesture',
                         'user' + '{0:0=2d}'.format(subj) + '*.aedat'))
        labelfile = glob.glob(
            os.path.join(datadir, 'DvsGesture',
                         'user' + '{0:0=2d}'.format(subj) + '*.csv'))
        datafile.sort()
        labelfile.sort()

        for idx in range(len(datafile)):
            dataset_tmp = get_gesture_data(datafile[idx], labelfile[idx])
            for data_dict in dataset_tmp:
                dataset.append(data_dict)
    return dataset


def download_and_extract_dataset(data_url, dest_directory):
    """Download and extract data set tar file.

    If the data set we're using doesn't already exist, this function
    downloads data from given url and unpacks it into a directory.
    If the data_url is none, don't download anything and expect the data
    directory to contain the correct files already.

    Args:
      data_url: Web location of the tar file containing the data set.
      dest_directory: File path to extract data to.
    """
    if not data_url:
        return
    if not os.path.exists(dest_directory):
        os.makedirs(dest_directory)
    filename = data_url.split('/')[-1]
    filepath = os.path.join(dest_directory, filename)
    if not os.path.exists(filepath):

        def _progress(count, block_size, total_size):
            sys.stdout.write('\r>> Downloading %s %.1f%%' %
                             (filename, float(count * block_size) /
                              float(total_size) * 100.0))
            sys.stdout.flush()

        try:
            filepath, _ = urllib.request.urlretrieve(data_url, filepath,
                                                     _progress)
        except:
            raise RuntimeError(
                f"Failed to download URL: {data_url} to folder: "
                f"{filepath}\nPlease make sure you have enough free"
                f" space and an internet connection")
        print()
        statinfo = os.stat(filepath)
        print(f"Successfully downloaded {filename} ({statinfo.st_size} bytes)")
    tarfile.open(filepath, 'r:gz').extractall(dest_directory)
